# import pickle as pkl

# train = pkl.load(open('./test_cls/meta.pkl', 'rb'))
# print(train)

# data = pkl.load(open('./test_cls/processed_data.pkl', 'rb'))
# print(data)



import pandas as pd
from dlk.utils.logger import setting_logger
setting_logger('process.log')
from dlk.utils.parser import config_parser_register
import copy
import json
import hjson
from dlk.data.processors import processor_config_register, processor_register

# train_data = json.load(open('./local_data/tasks_data/title/title_train.json', 'r'))
# valid_data = json.load(open('./local_data/tasks_data/title/title_valid.json', 'r'))
# test_data = json.load(open('./local_data/tasks_data/title/title_test.json', 'r'))


# NER benchmark data
# train_data = json.load(open('./local_data/tasks_data/ner_benchmark/train.json', 'r'))
# valid_data = json.load(open('./local_data/tasks_data/ner_benchmark/valid.json', 'r'))
# test_data = json.load(open('./local_data/tasks_data/ner_benchmark/test.json', 'r'))
train_data = json.load(open('./examples/sequence_labeling/conll2003/data/train.json', 'r'))
valid_data = json.load(open('./examples/sequence_labeling/conll2003/data/valid.json', 'r'))
test_data = json.load(open('./examples/sequence_labeling/conll2003/data/test.json', 'r'))
data = {"predict": test_data}
# data = {"test": test_data}

inp = {"data": data}
# config = config_parser_register.get("processor")(hjson.load(open("./examples/sequence_labeling/conll2003/norm_lstm_crf/prepro.hjson"),object_pairs_hook=dict)).parser_with_check()[0]
config = config_parser_register.get("processor")(hjson.load(open("./examples/sequence_labeling/conll2003/norm_char_lstm_crf/prepro.hjson"),object_pairs_hook=dict)).parser_with_check()[0]
config['data_dir'] = 'test_predict'
# config = config_parser_register.get("processor")(hjson.load(open("./examples/sequence_labeling/conll2003/norm_lstm_crf/pre_prepro.hjson"),object_pairs_hook=dict)).parser_with_check()[0]

# print(json.dumps(config, indent=4))
processor_register.get(config.get('_name'))(stage="predict", config=processor_config_register.get(config.get('_name'))(stage="train", config=config)).process(inp)



import sys
sys.path.append('../')
import pickle as pkl

# from dlk.utils.vocab import Vocabulary
# from dlk.models import DECODER_CONFIG_REGISTRY

# a = Vocabulary()
# a.auto_update(['ni', 'hao'])

# pkl.dump(a, open('vocab.pkl', 'wb'))

a = pkl.load(open('./vocab.pkl', 'rb'))
print(a.get_index('hao'))



"""
Provide BaseConfig which provide the basic method for configs, and ConfigTool a general config(dict) process tool
"""
from typing import Any, Dict, Union, Callable, List, Tuple, Type
import json
import copy
import os
from dlk.utils.logger import Logger
from dlk.utils.register import Register

logger = Logger.get_logger()

class BaseConfig(object):
    """docstring for BaseLayerConfig"""
    def __init__(self, config: Dict):
        super(BaseConfig, self).__init__()
        self._name = config.pop('_name')

    def post_check(self, config, used=[]):
        """check all the params are useful
        :config: TODO
        :returns: TODO
        """
        def rec_pop(cur_node, trace):
            """TODO: Docstring for rec_pop.
            """
            if len(trace) > 1:
                rec_pop(cur_node[trace[0]], trace[1:])
            if len(trace) == 1:
                if cur_node.get(trace[0], {}) == {}:
                    cur_node.pop(trace[0])

        config = copy.deepcopy(config)
        parant_traces = set()
        for key in used:
            sp_key = key.split('.')
            parant_traces.add(tuple(sp_key[:-1]))
            cur_root = config
            for key in sp_key[:-1]:
                cur_root = cur_root[key]
            cur_root.pop(sp_key[-1], None)
        for trace in parant_traces:
            rec_pop(config, trace)

        if config:
            logger.warning(f"In module '{self._name}', there are some params not be used: {config}")


class ConfigTool(object):
    """
    This Class is not be used as much as I design.
    """

    @staticmethod
    def _inplace_update_dict(_base, _new):
        """TODO: Docstring for _inplace_update_dict.
        :returns: TODO

        """
        for item in _new:
            if (item not in _base) or (not isinstance(_base[item], Dict)):
            # if item not in _base, or _base[item] is not Dict
                _base[item] = _new[item]
            elif isinstance(_base[item], Dict) and isinstance(_new[item], Dict):
                if "_name" in _base[item] and "_name" in _new[item]:
                    if _base[item]['_name'] != _new[item]['_name']:
                        logger.warning(f"The Higher Config for {_new[item]['_name']} Coverd the Base {_base[item]['_name']} ")
                        _base[item] = _new[item]

                        continue
                ConfigTool._inplace_update_dict(_base[item], _new[item])
            else:
                raise AttributeError("The base config and update config is not match. base: {}, new: {}. ".format(_base, _new))

    @staticmethod
    def do_update_config(config: dict, update_config: dict={}) ->Dict:
        """use update_config update the config

        :config: will updated config
        :returns: updated config

        """
        # BUG ?: if the config._name != update_config._name, should use the update_config conver the config wholely
        config = copy.deepcopy(config)
        ConfigTool._inplace_update_dict(config, update_config)
        return config

    @staticmethod
    def get_leaf_module(module_register: Register, module_config_register: Register, module_name: str, config: Dict) -> Tuple[Any, object]:
        """get sub module and config from register.
          for model, the leaf like encoder decoder and embedding class and the config of class could get by this mixin
        :module_register: Dict[model_name, Model]
        :module_config_register: Dict[model_config_name, ModelConfig]
        :module_name: for echo the log
        :config: Dict[key, value]
        :returns: tuple(Model, ModelConfig)

        """
        if isinstance(config, str):
            name = config
            extend_config = {}
        else:
            assert isinstance(config, dict), "{} config must be name(str) or config(dict), but you provide {}".format(module_name, config)
            name = config.get('_name', "") # must provide _name_
            extend_config = config
            if not name:
                raise KeyError('You must provide the {} name("name")'.format(module_name))

        module_class, module_config_class =  module_register.get(name), module_config_register.get(name)
        if (not module_class) or not (module_config_class):
            raise KeyError('The {} name {} is not registed.'.format(module_name, config))
        module_config = module_config_class(extend_config)
        return module_class, module_config

    @staticmethod
    def get_config_by_stage(stage:str, config:Dict)->Dict:
        """TODO: if config[stage] is a string, like 'train', 'predict' etc.,
            it means the config of this stage equals to config[stage]
            return config[config[stage]]
            e.g.
            config = {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "data_pair": {
                        "label": "label_id"
                    },
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'dev'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "vocab": "label_vocab", // usually provided by the "token_gather" module
                },
                "predict": "train",
                "online": ["train",
                {"vocab": "new_label_vocab"}
                ]
            }
            config.get_config['predict'] == config['predict'] == config['train']
        """
        config = config['config']
        stage_config = config.get(stage, {})
        if isinstance(stage_config, str):
            stage_config = config.get(stage_config, {})
        elif isinstance(stage_config, list):
            assert len(stage_config) == 2
            assert isinstance(stage_config[0], str)
            assert isinstance(stage_config[1], dict)
            stage_config = config.get(stage_config[0], {})
            ConfigTool.do_update_config(stage_config, stage_config[1])
        return stage_config



import torch.nn as nn
import torch
from typing import Dict, List
from . import module_register, module_config_register
from dlk.utils.config import BaseConfig


@module_config_register("linear")
class LinearConfig(BaseConfig):
    """docstring for LinearConfig
    {
        "config": {
            "input_size": 256,
            "output_size": 2,
            "dropout": 0.0, //the module output no need dropout
            "bias": true, // use bias or not in linear , if set to false, all the bias will be set to 0
            "pool": null, // pooling output or not
        },
        "_name": "linear",
    }
    """
    def __init__(self, config: Dict):
        super(LinearConfig, self).__init__(config)
        config = config['config']
        self.input_size = config['input_size']
        self.output_size = config['output_size']
        self.dropout = float(config['dropout'])
        self.bias = config['bias']
        self.pool = config['pool']
        self.post_check(config, used=[
            "input_size",
            "output_size",
            "dropout",
            "bias",
            "pool",
        ])


@module_register("linear")
class Linear(nn.Module):
    def __init__(self, config: LinearConfig):
        super(Linear, self).__init__()
        self.linear = nn.Linear(in_features=config.input_size, out_features=config.output_size, )
        self.dropout = nn.Dropout(p=float(config.dropout))
        self.config = config


    def forward(self, input: torch.Tensor)->torch.Tensor:
        """
        """
        output = self.dropout(self.linear(input))
        if not self.config.pool:
            return output
        elif self.config.pool == 'first':
            return output[:, 0]
        else:
            raise PermissionError(f"Currenttly we have not support the pool method '{self.config.pool}' in linear.")



import torch.nn as nn
from . import embedding_register, embedding_config_register
from typing import Callable, Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
import pickle as pkl
import torch


@embedding_config_register('combine_word_char_cnn')
class CombineWordCharCNNEmbeddingConfig(BaseModuleConfig):
    """docstring for BasicModelConfig, this exp. is for static word embedding and cnn char embedding
    {
        "_name": "combine_word_char_cnn",
        "embedding@char": {
            "_base": "static_char_cnn",
        },
        "embedding@word": {
            "_base": "static",
        },
        "config": {
            "word": {
                "embedding_file": "*@*", //the embedding file, must be saved as numpy array by pickle
                "embedding_dim": "*@*",
                "embedding_trace": ".", //default the file itself is the embedding
                "freeze": false, // is freeze
                "padding_idx": 0, //dropout rate
                "output_map": {"embedding": "word_embedding"},
                "input_map": {}, // required_key: provide_key
            },
            "char": {
                "embedding_file": "*@*", //the embedding file, must be saved as numpy array by pickle
                "embedding_dim": 35, //dropout rate
                "embedding_trace": ".", //default the file itself is the embedding
                "freeze": false, // is freeze
                "kernel_sizes": [3], //dropout rate
                "padding_idx": 0,
                "output_map": {"char_embedding": "char_embedding"},
                "input_map": {"char_ids": "char_ids"},
            },
            "dropout": 0, //dropout rate
            "embedding_dim": "*@*", // this must equal to char.embedding_dim + word.embedding_dim
            "output_map": {"embedding": "embedding"}, // this config do nothing, you can change this
            "input_map": {"char_embedding": "char_embedding", 'word_embedding': "word_embedding"}, // if the output of char and word embedding changed, you also should change this
        },
        "_link":{
            "config.word.embedding_file": ["embedding@word.config.embedding_file"],
            "config.word.embedding_dim": ["embedding@word.config.embedding_dim"],
            "config.word.embedding_trace": ["embedding@word.config.embedding_trace"],
            "config.word.freeze": ["embedding@word.config.freeze"],
            "config.word.padding_idx": ["embedding@word.config.padding_idx"],
            "config.word.output_map": ["embedding@word.config.output_map"],
            "config.word.input_map": ["embedding@word.config.input_map"],
            "config.char.embedding_file": ["embedding@char.config.embedding_file"],
            "config.char.embedding_dim": ["embedding@char.config.embedding_dim"],
            "config.char.embedding_trace": ["embedding@char.config.embedding_trace"],
            "config.char.freeze": ["embedding@char.config.freeze"],
            "config.char.kernel_sizes": ["embedding@char.config.kernel_sizes"],
            "config.char.padding_idx": ["embedding@char.config.padding_idx"],
            "config.char.output_map": ["embedding@char.config.output_map"],
            "config.char.input_map": ["embedding@char.config.input_map"],
        },
    }
    """
    def __init__(self, config: Dict):
        super(CombineWordCharCNNEmbeddingConfig, self).__init__(config)

        self.char_module_name = config['embedding@char']['_name']
        self.char_config = embedding_config_register[self.char_module_name](config['embedding@char'])
        self.word_module_name = config['embedding@word']['_name']
        self.word_config = embedding_config_register[self.word_module_name](config['embedding@word'])
        self.dropout = config['config']['dropout']
        self.embedding_dim = config['config']['embedding_dim']
        assert self.embedding_dim == self.char_config.embedding_dim + self.word_config.embedding_dim, f"The combine embedding_dim must equals to char_embedding+word_embedding, but {self.embedding_dim}!= {self.char_config.embedding_dim+self.word_config.embedding_dim}"

        self.post_check(config['config'], used=[
            "word.embedding_file",
            "word.embedding_dim",
            "word.embedding_trace",
            "word.freeze",
            "word.padding_idx",
            "word.output_map",
            "word.input_map",
            "char.embedding_file",
            "char.embedding_dim",
            "char.embedding_trace",
            "char.freeze",
            "char.padding_idx",
            "char.output_map",
            "char.input_map",
            "char.kernel_sizes",
            "dropout",
            "embedding_dim",
        ])


@embedding_register('combine_word_char_cnn')
class CombineWordCharCNNEmbedding(SimpleModule):
    """
    from 'input_ids' generate 'embedding'
    """

    def __init__(self, config: CombineWordCharCNNEmbeddingConfig):
        super().__init__(config)
        self._provided_keys = set() # provided by privous module, will update by the check_keys_are_provided
        self._provide_keys = {'char_embedding'} # provide by this module
        self._required_keys = {'char_ids'} # required by this module
        self.config = config
        self.dropout = nn.Dropout(float(self.config.dropout))
        self.word_embedding = embedding_register[self.config.word_module_name](self.config.word_config)
        self.char_embedding = embedding_register[self.config.char_module_name](self.config.char_config)

    def init_weight(self, method: Callable):
        """TODO: Docstring for init_weight.
        :arg1: TODO
        :returns: TODO
        """
        self.word_embedding.init_weight(method)
        self.char_embedding.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """forward
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """
        inputs = self.word_embedding(inputs)
        inputs = self.char_embedding(inputs)

        char_embedding = inputs[self.get_input_name("char_embedding")]
        word_embedding = inputs[self.get_input_name("word_embedding")]
        combine_embedding = torch.cat([char_embedding, word_embedding], dim=-1)
        inputs[self.get_output_name('embedding')] = self.dropout(combine_embedding)
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))
        return inputs



import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import ModelCheckpoint


@callback_config_register('checkpoint')
class CheckpointCallbackConfig(object):
    """docstring for CheckpointCallbackConfig
    {
        // default checkpoint configure
        "_name": "checkpoint",
        "config": {
            "monitor": "*@*",    // monitor which metrics or log value
            "save_top_k": 3,   //save top k
            "mode": "*@*", //"max" or "min" select topk min or max checkpoint, min for loss, max for acc
            "save_last": true,  //  always save last checkpoint
            "auto_insert_metric_name": true, //the save file name with or not metric name
            "every_n_train_steps": null, // Number of training steps between checkpoints.
            "every_n_epochs": 1, //Number of epochs between checkpoints.
            "save_on_train_epoch_end": false,// Whether to run checkpointing at the end of the training epoch. If this is False, then the check runs at the end of the validation.
            "save_weights_only": false, //whether save other status like optimizer, etc.
        }
    }
    """
    def __init__(self, config: Dict):
        super(CheckpointCallbackConfig, self).__init__()
        config = config['config']
        self.monitor = config['monitor']
        self.save_last = config['save_last']
        self.save_top_k = config['save_top_k']
        self.mode = config['mode']
        self.auto_insert_metric_name = config['auto_insert_metric_name']
        self.every_n_train_steps = config['every_n_train_steps']
        self.every_n_epochs = config['every_n_epochs']
        self.save_on_train_epoch_end = config['save_on_train_epoch_end']
        self.save_weights_only = config['save_weights_only']

@callback_register('checkpoint')
class CheckpointCallback(object):
    """
    """

    def __init__(self, config: CheckpointCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict):
        """get the checkpoint object
        :rt_config: Dict: runtime config, include save_dir, and the checkpoint path name
        :returns: checkpoint_callback object
        """
        dirpath = os.path.join(rt_config.get('save_dir', ''), rt_config.get("name", ''))
        return ModelCheckpoint(dirpath=dirpath, **self.config.__dict__)



"""registers"""
from typing import Callable, Dict, Any

class Register(object):
    """docstring for Register"""
    def __init__(self, register_name:str):
        super(Register, self).__init__()
        self.register_name = register_name
        self.registry:Dict[str, Any] = {}

    def register(self, name: str='')->Callable:
        """TODO: Docstring for register.
        :name: str: TODO
        :returns: TODO
        """
        def decorator(module):
            if name.strip() == "":
                raise ValueError(f'You must set a name for {module.__name__}')

            if name in self.registry:
                raise ValueError(f'The {name} is already registed in {self.register_name}.')
            self.registry[name] = module
            return module
        return decorator

    def __call__(self, name:str="")->Callable:
        """you can directly call the object, the behavior is the same as object.register(name)

        :name:str: TODO
        :returns: TODO

        """
        return self.register(name)

    def get(self, name: str='')->Any:
        """the name could be the real name or name+@+sub_name, and the
        :name: str: TODO
        :returns: TODO

        """
        sp_name = name.split('@')[0]
        if sp_name not in self.registry:
            raise KeyError(f"In '{self.register_name}' register, there is not a entry named '{sp_name}'")
        return self.registry[sp_name]

    def __getitem__(self, name: str='')->Any:
        """wrap for object.get(name)

        :arg1: TODO
        :returns: TODO
        """
        return self.get(name)



from logging import PercentStyle
import hjson
from typing import Dict, Union, Callable, List
import torch
from dlk.core.models import model_register, model_config_register
from dlk.core.optimizers import optimizer_register, optimizer_config_register
from dlk.core.schedulers import scheduler_register, scheduler_config_register
from dlk.core.losses import loss_register, loss_config_register
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register
from dlk.utils.config import BaseConfig, ConfigTool
from . import imodel_config_register, imodel_register, GatherOutputMixin
from dlk.utils.logger import Logger
from functools import lru_cache
logger = Logger.get_logger()

import pytorch_lightning as pl

@imodel_config_register('basic')
class BasicIModelConfig(BaseConfig):
    """ basic imodel config will provide all the config for model/optimizer/loss/scheduler/postprocess
    """
    def __init__(self, config: Dict):
        super(BasicIModelConfig, self).__init__(config)
        self.model, self.model_config = self.get_model(config.pop("model"))
        self.loss, self.loss_config = self.get_loss(config.pop("loss"))

        self.optimizer, self.optimizer_config = self.get_optimizer(config.pop("optimizer", 'adamw'))

        self.scheduler, self.scheduler_config = self.get_scheduler(config.pop("scheduler", "basic"))
        self.postprocess, self.postprocess_config = self.get_postprocessor(config.pop("postprocessor", 'identity'))

        self.config = config.pop('config', {})

    def get_postprocessor(self, config):
        """TODO: Docstring for get_postprocessor.
        :config: TODO
        :returns: TODO
        """
        return  ConfigTool.get_leaf_module(postprocessor_register, postprocessor_config_register, 'postprocessor', config)

    def get_model(self, config: Dict):
        """get embedding config and embedding module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(model_register, model_config_register, "model", config)

    def get_loss(self, config: Dict):
        """get encoder config and encoder module

        :config: TODO
        :returns: TODO

        """
        return ConfigTool.get_leaf_module(loss_register, loss_config_register, "loss", config)

    def get_optimizer(self, config: Dict):
        """get optimizer

        :config: TODO
        :returns: TODO
        """
        return ConfigTool.get_leaf_module(optimizer_register, optimizer_config_register, "optimizer", config)

    def get_scheduler(self, config: Dict):
        """get decoder config and decoder module

        :config: TODO
        :returns: TODO
        """
        return ConfigTool.get_leaf_module(scheduler_register, scheduler_config_register, "scheduler", config)


@imodel_register("basic")
class BasicIModel(pl.LightningModule, GatherOutputMixin):
    """
    """
    def __init__(self, config: BasicIModelConfig, checkpoint=False):
        """ init all modules except scheduler which requires the information from datamodule(training steps and every epoch steps)
        """
        super().__init__()
        self.config = config  # scheduler will init in configure_optimizers, because it will use the datamodule info

        self.model = config.model(config.model_config, checkpoint)


        self.calc_loss = config.loss(config.loss_config)
        self.get_optimizer = config.optimizer(model=self.model, config=config.optimizer_config)

        self._origin_valid_data = None
        self._origin_test_data = None
        self.postprocessor = config.postprocess(config.postprocess_config)
        self.gather_data = config.postprocess_config.input_map

    def get_progress_bar_dict(self):
        tqdm_dict = super().get_progress_bar_dict()
        tqdm_dict.pop("v_num", None)
        return tqdm_dict

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        return self.model(inputs)

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int):
        result = self.model.training_step(batch)
        loss = self.calc_loss(result, batch, rt_config={
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        self.log_dict({"train_loss": loss.unsqueeze(0)}, prog_bar=True)
        return loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int):
        result = self.model.validation_step(batch)
        loss = self.calc_loss(result, batch, rt_config={  # align with training step
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        gather_column = list(self.gather_data.keys())
        return_result = {"loss": loss.unsqueeze(0)} # this loss will use in postprocess
        for column in gather_column:
            if column in result:
                return_result[column] = result[column]
        return_result['_index'] = batch['_index']
        return return_result

    def validation_epoch_end(self, outputs):
        """TODO: Docstring for test_epoch_end.
        :returns: TODO
        """
        outputs = self.gather_outputs(outputs)

        self.log_dict(
            self.postprocessor(stage='valid', list_batch_outputs=outputs, origin_data=self._origin_valid_data,
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs
                }),
            prog_bar=True, rank_zero_only=True)
        return outputs

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int):
        result = self.model.test_step(batch)
        loss = self.calc_loss(result, batch, rt_config={  # align with training step
            "current_step": self.global_step,
            "current_epoch": self.current_epoch,
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })
        gather_column = list(self.gather_data.keys())
        return_result = {"loss": loss.unsqueeze(0)} # this loss will use in postprocess
        for column in gather_column:
            if column in result:
                return_result[column] = result[column]
        return_result['_index'] = batch['_index']
        return return_result

    def test_epoch_end(self, outputs):
        """TODO: Docstring for test_epoch_end.
        :returns: TODO

        """
        outputs = self.gather_outputs(outputs)

        # if self.local_rank in [0, -1]:
        self.log_dict(
            self.postprocessor(stage='test', list_batch_outputs=outputs, origin_data=self._origin_test_data,
                rt_config={
                    "current_step": self.global_step,
                    "current_epoch": self.current_epoch,
                    "total_steps": self.num_training_steps,
                    "total_epochs": self.num_training_epochs
                }),
            prog_bar=True, rank_zero_only=True)
        return outputs

    def predict_step(self, batch, batch_idx):
        return self.model.predict_step(batch)

    @property
    @lru_cache(maxsize=5) # the size should always == 1
    def num_training_epochs(self) -> int:
        """Total training epochs inferred from datamodule and devices.
        """
        return self.trainer.max_epochs

    @property
    @lru_cache(maxsize=5) # the size should always == 1
    def epoch_training_steps(self) -> int:
        """every epoch training steps inferred from datamodule and devices.
        """
        if self.trainer.datamodule.train_dataloader() is None:
            batches = 0
        else:
            batches = len(self.trainer.datamodule.train_dataloader())
        return batches

    @property
    @lru_cache(maxsize=5) # the size should always == 1
    def num_training_steps(self) -> int:
        """Total training steps inferred from datamodule and devices.
        """
         # TODO: https://github.com/PyTorchLightning/pytorch-lightning/issues/5449 should check update
        if self.trainer.max_steps != -1:
            return self.trainer.max_steps

        limit_batches = self.trainer.limit_train_batches
        if self.trainer.datamodule.train_dataloader() is None:
            batches = 0
        else:
            batches = len(self.trainer.datamodule.train_dataloader())
        batches = min(batches, limit_batches) if isinstance(limit_batches, int) else int(limit_batches * batches)

        num_devices = max(1, self.trainer.num_gpus, self.trainer.num_processes)
        if self.trainer.tpu_cores:
            num_devices = max(num_devices, self.trainer.tpu_cores)

        effective_accum = self.trainer.accumulate_grad_batches * num_devices
        return (batches // effective_accum + (1 if batches%effective_accum else 0)) * self.trainer.max_epochs

    def configure_optimizers(self):

        self.calc_loss.update_config(rt_config={
            "total_steps": self.num_training_steps,
            "total_epochs": self.num_training_epochs
        })

        optimizer = self.get_optimizer()
        if "num_training_steps" in self.config.scheduler_config.__dict__:
            self.config.scheduler_config.num_training_steps = self.num_training_steps
        if "epoch_training_steps" in self.config.scheduler_config.__dict__:
            self.config.scheduler_config.epoch_training_steps = self.epoch_training_steps
        if "num_training_epochs" in self.config.scheduler_config.__dict__:
            self.config.scheduler_config.num_training_epochs = self.num_training_epochs
        # self.config.scheduler_config.last_epoch = -1
        scheduler = self.config.scheduler(optimizer, self.config.scheduler_config)()

        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": scheduler,
                "interval": "step",
                "frequency": 1
            }
        }



"""
Loader the $meta, etc. to data
"""
from dlk.utils.config import ConfigTool, BaseConfig
from dlk.utils.logger import Logger
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
import pickle as pkl
import os

logger = Logger.get_logger()

@subprocessor_config_register('load')
class LoadConfig(BaseConfig):
    """
    Config eg.
     {
         "_name": "load",
         "config":{
             "base_dir": "."
             "predict":{
                 "meta": "./meta.pkl",
             },
             "online": [
                 "predict", //base predict
                 {   // special config, update predict, is this case, the config is null, means use all config from "predict", when this is empty dict, you can only set the value to a str "predict", they will get the same result
                 }
             ]
         }
     },
    """

    def __init__(self, stage, config):
        super(LoadConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.base_dir:str = config.get('config').get("base_dir", ".")


@subprocessor_register('load')
class Load(ISubProcessor):
    """
    """

    def __init__(self, stage: str, config: LoadConfig):
        super().__init__()
        self.stage = stage
        self.config = config.config
        if not self.config:
            logger.info(f"Skip 'load' at stage {self.stage}")
            return
        self.base_dir = config.base_dir

    def load(self, path):
        """TODO: Docstring for load.
        """

        logger.info(f"Loading file from {os.path.join(self.base_dir, path)}")
        return pkl.load(open(os.path.join(self.base_dir, path), 'rb'))

    def process(self, data: Dict)->Dict:
        for _, path in self.config.items():
            meta = self.load(path)
            for key, value in meta.items():
                data[key] = value
        return data



"""embeddings"""
import importlib
import os
from typing import Callable, Dict, Tuple, Any
from dlk.utils.register import Register

embedding_config_register = Register("Embedding config register.")
embedding_register = Register("Embedding register.")


def import_embeddings(embeddings_dir, namespace):
    for file in os.listdir(embeddings_dir):
        path = os.path.join(embeddings_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            embedding_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + embedding_name)


class EmbeddingInput(object):
    """docstring for EmbeddingInput"""
    def __init__(self, **args):
        super(EmbeddingInput, self).__init__()
        self.input_ids = args.get("input_ids", None)


class EmbeddingOutput(object):
    """docstring for EmbeddingOutput"""
    def __init__(self, **args):
        super(EmbeddingOutput, self).__init__()
        self.represent = args.get("represent", None)


# automatically import any Python files in the embeddings directory
embeddings_dir = os.path.dirname(__file__)
import_embeddings(embeddings_dir, "dlk.core.layers.embeddings")



import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import ProgressBarBase


@callback_config_register('progressbar')
class ProgressBarCallbackConfig(object):
    """
    {
        // default progressbar configure
        "_name": progressbar,
        "config": {
            "disable_v_num": true,
        }
    }
    """
    def __init__(self, config):
        super(ProgressBarCallbackConfig, self).__init__()
        config = config['config']
        self.disable_v_num = config['disable_v_num']



@callback_register('progressbar')
class ProgressBarCallback(object):
    """
    """

    def __init__(self, config: ProgressBarCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict):
        """TODO: Docstring for __call__.

        :rt_config: Dict: TODO
        :returns: TODO

        """
        raise PermissionError(f"Don't use Progressbar callback, it's not implemented yet.")
        return ProgressBarBase()



import ahocorasick
from typing import Iterable, List, Dict

class QuickSearch(object):
    """docstring for QuickSearch"""
    def __init__(self, words: Iterable=[]):
        """
            words: use words to init the QuickSearch
        """
        super(QuickSearch, self).__init__()

        if not isinstance(words, Iterable) :
            raise PermissionError("The words must be a Iterable object.")
        self.next_index = 0

        self.ac = ahocorasick.Automaton()

        for sub_str in words:
            sub_str = f" {sub_str.strip()} "
            self.ac.add_word(sub_str, (self.next_index, sub_str))
            self.next_index += 1
        self.ac.make_automaton()

    def search(self, search_str:str)->List[Dict]:
        """find whether some sub_str in QuickSearch
        :search_str: TODO
        :returns: List of result
            the result organized as {
                "start": start_position,
                "end": end_position,
                "str": search_str[start_position: end_position]
            }
        """
        result = []
        for end_index, (_, original_value) in self.ac.iter(search_str):
            start_index = end_index - len(original_value) + 1
            end_index = end_index + 1
            result.append({"start": start_index, "end": end_index, "str": original_value})
        return result

    def has(self, key: str)->bool:
        """find key is in our data
        :key: str
        :returns: true|false

        """
        try:
            self.ac.get(key)
        except:
            return False
        return True

    def add_word(self, word: str):
        """add a single word to QuickSearch
        :word: str
        :returns: None
        """
        self.ac.add_word(word, (self.next_index, word))
        self.next_index += 1
        self.ac.make_automaton()

    def add_words(self, words: Iterable=[]):
        """add words from iterator to the ac
        :words: Iterable
        :returns: None
        """
        for sub_str in words:
            self.ac.add_word(sub_str, (self.next_index, sub_str))
            self.next_index += 1
        self.ac.make_automaton()



import hjson
from tokenizers import Tokenizer
from typing import Dict
import argparse
import json
from transformers import RobertaTokenizerFast, BertTokenizerFast
# import os
# os.environ["TOKENIZERS_PARALLELISM"] = "false"


TOKENIZER_MAP = {
    "bert": BertTokenizerFast,
    "roberta": RobertaTokenizerFast
}


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default='./convert.hjson',
        help="The config path.",
    )

    parser.add_argument(
        "--type", type=str, default='roberta', help="roberta, bert", choices=['roberta', "bert"]
    )

    parser.add_argument("--config_dir", type=str, default='./roberta/', help="transformers tokenizer config files")
    parser.add_argument("--output", type=str, default='./roberta_tokenizer.json', help="transformers tokenizer config files")


    args = parser.parse_args()
    if args.config:
        config_json = hjson.load(open(args.config), object_pairs_hook=dict)
        for key, value in config_json.items():
           setattr(args, key, value)


    tokenizer = TOKENIZER_MAP[args.type].from_pretrained(args.config_dir)
    # print(len(tokenizer._tokenizer.get_vocab()))
    str = tokenizer._tokenizer.to_str()
    # json.dump(json.loads(str), open(args.output, 'w'), indent=4)

    tokenizer._tokenizer.save(pretty=True, path=args.output)



"""datamodules"""

import importlib
import os
from typing import Dict, Any
from dlk.utils.register import Register
from pytorch_lightning import LightningDataModule
import abc
from torch.nn.utils.rnn import pad_sequence
import torch

datamodule_config_register = Register("Datamodule config register")
datamodule_register = Register("Datamodule register")


collate_register = Register('Collate function register.')



@collate_register('default')
class DefaultCollate(object):
    """docstring for DefaultCollate"""
    def __init__(self, **config):
        super(DefaultCollate, self).__init__()
        self.key_padding_pairs = config.get("key_padding_pairs", {})
        self.gen_mask = config.get("gen_mask", {})

    def __call__(self, batch):
        keys = batch[0].keys()
        data_map: Dict[str, Any] = {}
        for key in keys:
            data_map[key] = []
        for key in keys:
            for one_ins in batch:
                data_map[key].append(one_ins[key])
        if self.gen_mask:
            for key, mask in self.gen_mask.items():
                data_map[mask] = []
                for item in data_map[key]:
                    data_map[mask].append(torch.tensor([1] * len(item), dtype=torch.int))
        for key in data_map:
            try:
                data_map[key] = pad_sequence(data_map[key], batch_first=True, padding_value=self.key_padding_pairs.get(key, 0))
            except:
                # if the data_map[key] is size 0, we can concat them
                if data_map[key][0].size():
                    raise ValueError(f"The {data_map[key]} can not be concat by pad_sequence.")
                _data = pad_sequence([i.unsqueeze(0) for i in data_map[key]], batch_first=True, padding_value=self.key_padding_pairs.get(key, 0)).squeeze()
                if not _data.size():
                    _data.unsqueeze_(0)
                data_map[key] = _data
        return data_map


class IBaseDataModule(LightningDataModule):
    """docstring for IBaseDataModule"""
    def __init__(self):
        super(IBaseDataModule, self).__init__()

    def train_dataloader(self):
        raise NotImplementedError(f"You must implementation the train_dataloader for your own datamodule.")

    def predict_dataloader(self):
        raise NotImplementedError(f"You must implementation the predict_dataloader for your own datamodule.")

    def val_dataloader(self):
        raise NotImplementedError(f"You must implementation the val_dataloader for your own datamodule.")

    def test_dataloader(self):
        raise NotImplementedError(f"You must implementation the test_dataloader for your own datamodule.")

    @abc.abstractmethod
    def online_dataloader(self):
        raise NotImplementedError(f"You must implementation the online_dataloader for your own datamodule.")


def import_datamodules(datamodules_dir, namespace):
    for file in os.listdir(datamodules_dir):
        path = os.path.join(datamodules_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            datamodule_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + datamodule_name)


# automatically import any Python files in the models directory
datamodules_dir = os.path.dirname(__file__)
import_datamodules(datamodules_dir, "dlk.data.datamodules")



import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import BaseModule, BaseModuleConfig
from . import decoder_register, decoder_config_register
from dlk.core.modules import module_config_register, module_register
import copy

@decoder_config_register("linear_crf")
class LinearCRFConfig(BaseModuleConfig):
    """docstring for LinearCRFConfig
    {
        "module@linear": {
            _base: linear,
        },
        "module@crf": {
            _base: crf,
        },
        "config": {
            "input_size": "*@*",
            "output_size": "*@*",
            "reduction": "mean",
            "output_map": {}, //provide_key: output_key
            "input_map": {} // required_key: provide_key
        },
        "_link":{
            "config.input_size": ["module@linear.config.input_size"],
            "config.output_size": ["module@linear.config.output_size", "module@crf.config.output_size"],
            "config.reduction": ["module@crf.config.reduction"],
        }
        "_name": "linear_crf",
    }
    """
    def __init__(self, config: Dict):
        super(LinearCRFConfig, self).__init__(config)
        self.linear_config = config["module@linear"]
        self.crf_config = config["module@crf"]
        self.post_check(config['config'], used=['input_size', 'output_size', 'reduction'])


@decoder_register("linear_crf")
class LinearCRF(BaseModule):
    def __init__(self, config: LinearCRFConfig):
        super(LinearCRF, self).__init__(config)
        self._provide_keys = {'logits', "predict_seq_label"}
        self._required_keys = {'embedding', 'label_ids', 'attention_mask'}

        self.config = config
        self.linear = module_register.get('linear')(module_config_register.get('linear')(config.linear_config))
        self.crf = module_register.get('crf')(module_config_register.get('crf')(config.crf_config))

    def init_weight(self, method: Callable):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        for module in self.linear.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """
        """
        return self.predict_step(inputs)

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """predict
        :inputs: Dict[str: torch.Tensor], one mini-batch inputs
        :returns: Dict[str: torch.Tensor], one mini-batch outputs
        """

        logits = self.linear(inputs[self.get_input_name('embedding')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name("predict_seq_label")] = self.crf(logits, inputs[self.get_input_name('attention_mask')])
        return inputs

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """TODO: Docstring for training_step.
        :arg1: TODO
        :returns: TODO

        """
        logits = self.linear(inputs[self.get_input_name('embedding')])
        loss = self.crf.training_step(logits, inputs[self.get_input_name('label_ids')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name('loss')] = loss
        return inputs

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """TODO: Docstring for training_step.

        :arg1: TODO
        :returns: TODO

        """
        logits = self.linear(inputs[self.get_input_name('embedding')])
        loss = self.crf.training_step(logits, inputs[self.get_input_name('label_ids')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([logits]))
        inputs[self.get_output_name('loss')] = loss
        inputs[self.get_output_name("predict_seq_label")] = self.crf(logits, inputs[self.get_input_name('attention_mask')])
        return inputs



"""postprocessors"""

import importlib
import os
from typing import Callable, Dict, Type
from dlk.utils.register import Register
from dlk.utils.config import BaseConfig
import abc

class IPostProcessorConfig(BaseConfig):
    """docstring for PostProcessorConfigBase"""
    def __init__(self, config):
        super(IPostProcessorConfig, self).__init__(config)
        self.config = config.get('config', {})

    @property
    def input_map(self):
        """required the output of model process content name map
        :returns: TODO

        """
        return self.config.get("input_map", {})

    @property
    def origin_input_map(self):
        """required the origin data(before pass to datamodule) column name map
        :returns: TODO

        """
        return self.config.get("origin_input_map", {})


class IPostProcessor(metaclass=abc.ABCMeta):
    """docstring for IPostProcessor"""

    def loss_name_map(self, stage):
        """TODO: Docstring for loss_name_map.
        :returns: TODO

        """
        map = {
            "valid": 'val',
            'train': 'train',
            "test": "test",
        }
        return map.get(stage, stage)

    def average_loss(self, list_batch_outputs):
        """TODO: Docstring for average_loss.

        :list_batch_outputs: TODO
        :returns: TODO

        """
        sum_loss = 0
        for batch_output in list_batch_outputs:
            sum_loss += batch_output.get('loss', 0)
        return sum_loss / len(list_batch_outputs)

    @abc.abstractmethod
    def do_predict(self, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for do_predict.
        :stage: TODO
        :list_batch_outputs: TODO
        :origin_data: TODO
        :rt_config: TODO
        :returns: TODO

        """
        raise NotImplementedError

    @abc.abstractmethod
    def do_calc_metrics(self, predicts, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for do_calc_metrics.
        :returns: TODO

        """
        raise NotImplementedError

    @abc.abstractmethod
    def do_save(self, predicts, stage, list_batch_outputs, origin_data, rt_config={}, save_condition=False):
        """TODO: Docstring for do_save.

        :predicts: TODO
        :rt_config: TODO
        :condition: when the save condition is True, do save
        :returns: TODO
        """
        raise NotImplementedError

    @property
    def without_ground_truth_stage(self):
        """TODO: Docstring for no_ground_truth_stage.
        :returns: TODO

        """
        return {'predict', 'online'}

    def process(self, stage, list_batch_outputs, origin_data, rt_config)->Dict:
        """TODO: Docstring for process.

        :arg1: TODO
        :returns: TODO

        """
        log_info = {}
        if stage not in self.without_ground_truth_stage:
            average_loss = self.average_loss(list_batch_outputs=list_batch_outputs)
            log_info[f'{self.loss_name_map(stage)}_loss'] = average_loss
        predicts = self.do_predict(stage, list_batch_outputs, origin_data, rt_config)
        if stage not in self.without_ground_truth_stage:
            log_info.update(self.do_calc_metrics(predicts, stage, list_batch_outputs, origin_data, rt_config))
        if stage == 'online':
            return predicts
        if stage == 'predict':
            self.do_save(predicts, stage, list_batch_outputs, origin_data, rt_config, save_condition=True)
        else:
            self.do_save(predicts, stage, list_batch_outputs, origin_data, rt_config, save_condition=False)
        return log_info

    def __call__(self, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for __call.
        :returns: TODO

        """
        return self.process(stage, list_batch_outputs, origin_data, rt_config)


postprocessor_config_register = Register('PostProcessor config register')
postprocessor_register = Register("PostProcessor register")


def import_postprocessors(postprocessors_dir, namespace):
    for file in os.listdir(postprocessors_dir):
        path = os.path.join(postprocessors_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            # and not (file.endswith("subpostprocessors") and os.path.isdir(path))
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            postprocessor_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + postprocessor_name)


# automatically import any Python files in the models directory
postprocessors_dir = os.path.dirname(__file__)
import_postprocessors(postprocessors_dir, "dlk.data.postprocessors")



from dlk.utils.logger import setting_logger
import pytorch_lightning as pl
# setting_logger("./benchmark.log")
# setting_logger("./conll_norm_lstm_crf.log")
setting_logger("./conll_norm_char_lstm_crf.log")
from dlk.predict import Predict
import json

pl.seed_everything(88)

# trainer = Train('./examples/sequence_labeling/benchmark/pretrained/first_piece_lstm_crf_main.hjson')
predictor = Predict('./examples/sequence_labeling/conll2003/norm_char_lstm_crf/output/task=crf_lstm_lr=0.01_optimizer=sgd_dropout=0.5_batch_size=10_lstm_output_size=200/config.json', './examples/sequence_labeling/conll2003/norm_char_lstm_crf/output/task=crf_lstm_lr=0.01_optimizer=sgd_dropout=0.5_batch_size=10_lstm_output_size=200/default/checkpoints/epoch=0-step=1404.ckpt')
# trainer = Train('./examples/sequence_labeling/conll2003/norm_lstm_crf/main.hjson')
predictor.predict()

# trainer.run()



"""
Get the dlk package root path
"""
import os

ROOT = os.path.join('/')

def get_root():
    """TODO: Docstring for get_root.
    :returns: TODO

    """
    cur_dir = os.path.dirname(os.path.abspath(__file__))

    def check_cur_is_root(cur_dir):
        """TODO: Docstring for check_cur_is_root.
        :arg1: TODO
        :returns: TODO
        """
        root_sign_list = ['dlk']
        for root_sign in root_sign_list:
            if os.path.exists(os.path.join(cur_dir, root_sign)):
                return True
        return False

    while cur_dir != ROOT:
        if check_cur_is_root(cur_dir):
            break
        cur_dir = os.path.dirname(cur_dir)

    if cur_dir == ROOT:
        raise ImportError(f'{os.path.dirname(os.path.abspath(__file__))} is not a right file to find dlk root')
    else:
        return cur_dir



import torch.nn as nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from typing import Dict, List, Set
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import embedding_register, embedding_config_register
from dlk.core.modules import module_config_register, module_register

@embedding_config_register("pretrained_transformers")
class PretrainedTransformersConfig(BaseModuleConfig):
    """docstring for PretrainedTransformersConfig
    {
        "module": {
            "_base": "roberta",
        },
        "config": {
            "pretrained_model_path": "*@*",
            "input_map": {
                "input_ids": "input_ids",
                "attention_mask": "attention_mask",
                "type_ids": "type_ids",
            },
            "output_map": {
                "embedding": "embedding",
            },
            "dropout": 0, //dropout rate
            "embedding_dim": "*@*",
        },
        "_link": {
            "config.pretrained_model_path": ["module.config.pretrained_model_path"],
        },
        "_name": "pretrained_transformers",
    }
    for gather embedding
    {
        "module": {
            "_base": "roberta",
        },
        "config": {
            "pretrained_model_path": "*@*",
            "input_map": {
                "input_ids": "input_ids",
                "attention_mask": "subword_mask",
                "type_ids": "type_ids",
                "gather_index": "gather_index",
            },
            "output_map": {
                "embedding": "embedding",
            },
            "embedding_dim": "*@*",
            "dropout": 0, //dropout rate
        },
        "_link": {
            "config.pretrained_model_path": ["module.config.pretrained_model_path"],
        },
        "_name": "pretrained_transformers",
    }
    """

    def __init__(self, config: Dict):
        super(PretrainedTransformersConfig, self).__init__(config)
        self.pretrained_transformers_config = config["module"]
        self.post_check(config['config'], used=[
            "pretrained_model_path",
            "embedding_dim",
            "output_map",
            "input_map",
            "dropout",
            ])


@embedding_register("pretrained_transformers")
class PretrainedTransformers(SimpleModule):
    def __init__(self, config: PretrainedTransformersConfig):
        super(PretrainedTransformers, self).__init__(config)
        self._provide_keys = {'embedding'}
        self._required_keys = {'input_ids', 'attention_mask'}
        self.config = config
        self.pretrained_transformers = module_register.get(config.pretrained_transformers_config['_name'])(module_config_register.get(config.pretrained_transformers_config['_name'])(config.pretrained_transformers_config))

    def init_weight(self, method):
        """init  Module weight by `method`
        :method: init method, with pretrained
        :returns: None
        """
        self.pretrained_transformers.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """
        """
        input_ids = inputs[self.get_input_name('input_ids')] if "input_ids" in self.config._input_map else None
        attention_mask = inputs[self.get_input_name('attention_mask')] if "attention_mask" in self.config._input_map else None
        type_ids = inputs[self.get_input_name('type_ids')] if "type_ids" in self.config._input_map else None
        type_ids = inputs[self.get_input_name('type_ids')] if "type_ids" in self.config._input_map else None
        inputs_embeds = inputs[self.get_input_name('inputs_embeds')] if "inputs_embeds" in self.config._input_map else None
        if (input_ids is None and inputs_embeds is None) or (input_ids is not None and inputs_embeds is not None):
            raise PermissionError("input_ids and input_embeds must set one of them to None")
        sequence_output, all_hidden_states, all_self_attentions = self.pretrained_transformers(
            {
                "input_ids": input_ids,
                "attention_mask": attention_mask,
                "token_type_ids": type_ids,
                "inputs_embeds": inputs_embeds,
            }
        )
        if 'gather_index' in self.config._input_map:
            # gather_index.shape == bs*real_sent_len
            gather_index = inputs[self.get_input_name("gather_index")]
            g_bs, g_seq_len = gather_index.shape
            bs, seq_len, hid_size = sequence_output.shape
            assert g_bs == bs
            assert g_seq_len <= seq_len
            sequence_output = torch.gather(sequence_output[:, :, :], 1, gather_index.unsqueeze(-1).expand(bs, g_seq_len, hid_size))
        inputs[self.get_output_name('embedding')] = sequence_output
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather(all_hidden_states))
        return inputs



import hjson
import os
from typing import Dict, Union, Callable, List, Any
# from models import MODEL_REGISTRY, MODEL_CONFIG_REGISTRY
from dlk.utils.parser import CONFIG_PARSER_REGISTRY
import json

#TODO: parser应该返回完整的参数

def load_hjson_file(file_name: str) -> Dict:
    """load hjson file by file_name

    :file_name: TODO
    :returns: TODO

    """
    json_file = hjson.load(open(file_name), object_pairs_hook=dict)
    return json_file

def test_base():
    """TODO: Docstring for test_base.
    :returns: TODO

    """
    config_file = load_hjson_file('./tests/parser/test_base_inp.hjson')
    out = [json.dumps(conf, sort_keys=True) for conf in load_hjson_file('./tests/parser/test_base_out.hjson')['out']]
    out.sort()
    focus = config_file.pop('_focus', {})
    parser = CONFIG_PARSER_REGISTRY['task'](config_file)
    configs = parser.parser()
    config_names = []

    for possible_config in configs:
        config_name = []
        if not focus:
            config_name.append(str(possible_config))
        else:
            for source, to in focus.items():
                config_point = possible_config
                trace = source.split('.')
                for t in trace:
                    config_point = config_point[t]
                config_name.append(to+str(config_point))
        config_names.append('_'.join(config_name))
    assert len(config_names) == len(set(config_names))
    for conf in configs:
        print(json.dumps(conf, indent=4))
    configs = [json.dumps(conf, sort_keys=True) for conf in configs]
    configs.sort()
    assert configs == out
    # if len(config_names) != len(set(config_names)):
        # print(len(config_names))
        # print(len(set(config_names)))
        # for config, name in zip(configs, config_names):
            # print(json.dumps(config, indent=4))
            # print(name)
        # raise NameError('The config_names is not unique.')
    # for config, name in zip(configs, config_names):
        # print(json.dumps(config, indent=4))
        # print(name)

test_base()
# Train('simple_ner')
# Train('./dlk/configures/tasks/simple_ner.hjson')
# Train('lstm_linear_ner')




import json
import hjson
import argparse

pre_tokenizers = {
    "bert": {"type": "BertPreTokenizer"},
    "whitespacesplit": {"type": "WhitespaceSplit"}
}

template = {
    "version": "1.0",
    "truncation": None,
    "padding": None,
    "added_tokens": [
        {
            "id": 0,
            "special": True,
            "content": "[CLS]",
            "single_word": False,
            "lstrip": False,
            "rstrip": False,
            "normalized": False
        },
    ],
    "normalizer": {
        "type": "BertNormalizer",
        "clean_text": True,
        "handle_chinese_chars": True,
        "strip_accents": None,
        "lowercase": True
    },
    "pre_tokenizer": {
        "type": "BertPreTokenizer"
    },
    "post_processor": {
        "type": "TemplateProcessing",
        "single": [
            {
                "SpecialToken": {
                    "id": "[CLS]",
                    "type_id": 0
                }
            },
            {
                "Sequence": {
                    "id": "A",
                    "type_id": 0
                }
            },
            {
                "SpecialToken": {
                    "id": "[SEP]",
                    "type_id": 0
                }
            }
        ],
        "pair": [
            {
                "SpecialToken": {
                    "id": "[CLS]",
                    "type_id": 0
                }
            },
            {
                "Sequence": {
                    "id": "A",
                    "type_id": 0
                }
            },
            {
                "SpecialToken": {
                    "id": "[SEP]",
                    "type_id": 0
                }
            },
            {
                "Sequence": {
                    "id": "B",
                    "type_id": 1
                }
            },
            {
                "SpecialToken": {
                    "id": "[SEP]",
                    "type_id": 1
                }
            }
        ],
        "special_tokens": {
            "[CLS]": {
                "id": "[CLS]",
                "ids": [
                    0
                ],
                "tokens": [
                    "[CLS]"
                ]
            },
            "[SEP]": {
                "id": "[SEP]",
                "ids": [
                    1
                ],
                "tokens": [
                    "[SEP]"
                ]
            }
        }
    },
    "decoder": {
        "type": "WordPiece",
        "prefix": "##",
        "cleanup": True
    },
    "model": {
        "type": "WordPiece",
        "unk_token": "[UNK]",
        "continuing_subword_prefix": "##",
        "max_input_chars_per_word": 100,
        "vocab": {
        }
    }
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        type=str,
        default='./vocab2tokenizer.hjson',
        help="The config path.",
    )

    parser.add_argument(
        "--unk",
        type=str,
        default='[UNK]',
        help="The unk token repretation.",
    )

    parser.add_argument(
        "--vocab",
        type=str,
        default='vocab.txt',
        help="The vocab path.",
    )

    parser.add_argument(
        "--do_norm",
        type=bool,
        default=True,
        help="Do BertNormalizer or not.",
    )

    parser.add_argument(
        "--do_lowercase",
        type=bool,
        default=True,
        help="Normalize do lowercase or not.",
    )
    parser.add_argument(
        "--pre_tokenizer",
        type=str,
        default='bert',
        help="bert or whitespacesplit",
    )

    parser.add_argument(
        "--do_bert_postprocess",
        type=bool,
        default=True,
        help="Whether add the [CLS] and [SEP] token for single or pair input.",
    )

    parser.add_argument("--output", type=str, default='./vocab_tokenizer.json', help="generated transformers tokenizer config files")


    args = parser.parse_args()
    if args.config:
        config_json = hjson.load(open(args.config), object_pairs_hook=dict)
        for key, value in config_json.items():
           setattr(args, key, value)

    tokenizer = template

    added_tokens = ['[CLS]', '[SEP]', '[MASK]', "[PAD]"] + [args.unk]

    added_tokens_config = []
    vocab = {}
    for i, token in enumerate(added_tokens):
        added_templete = {
              "id": 0,
              "special": True,
              "content": "",
              "single_word": True,
              "lstrip": False,
              "rstrip": False,
              "normalized": False
        }
        added_templete['id'] = i
        added_templete['content'] = token
        added_tokens_config.append(added_templete)
        if token not in vocab:
            vocab[token] = i

    normalizer = {
        "type": "BertNormalizer",
        "clean_text": True,
        "handle_chinese_chars": True,
        "strip_accents": None,
        "lowercase": True if args.do_lowercase else False
    }

    with open(args.vocab, 'r') as f:
        lines = f.readlines()
        for line in lines:
            token = line.strip()
            if token not in vocab:
                vocab[token] = len(vocab)
    tokenizer['added_tokens'] = added_tokens_config
    tokenizer['model']['vocab'] = vocab
    tokenizer['model']['unk_token'] = args.unk
    tokenizer['normalizer'] = normalizer if args.do_norm else None
    tokenizer['pre_tokenizer'] = pre_tokenizers[args.pre_tokenizer]
    if not args.do_bert_postprocess:
        tokenizer['post_processor'] = None

    json.dump(tokenizer, open(args.output, 'w'), indent=4)



import pickle as pkl
import json
from typing import Dict, List, Optional, Tuple, Union
import os
import numpy as np
from typing import Dict
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.logger import Logger
from dlk.utils.vocab import Vocabulary
from tokenizers import Tokenizer
import torchmetrics
logger = Logger.get_logger()


@postprocessor_config_register('seq_lab')
class SeqLabPostProcessorConfig(IPostProcessorConfig):
    """docstring for SeqLabPostProcessorConfig
    config e.g.
    {
        "_name": "seq_lab",
        "config": {
            "meta": "*@*",
            "use_crf": false, //use or not use crf
            "word_ready": false, //already gather the subword first token as the word rep or not
            "meta_data": {
                "label_vocab": 'label_vocab',
                "tokenizer": "tokenizer",
            },
            "input_map": {
                "logits": "logits",
                "predict_seq_label": "predict_seq_label",
                "_index": "_index",
            },
            "origin_input_map": {
                "uuid": "uuid",
                "sentence": "sentence",
                "input_ids": "input_ids",
                "entities_info": "entities_info",
                "offsets": "offsets",
                "special_tokens_mask": "special_tokens_mask",
                "word_ids": "word_ids",
                "label_ids": "label_ids",
            },
            "save_root_path": ".",  //save data root dir
            "save_path": {
                "valid": "valid",  // relative dir for valid stage
                "test": "test",    // relative dir for test stage
            },
            "start_save_step": 0,  // -1 means the last
            "start_save_epoch": -1,
            "aggregation_strategy": "max", // AggregationStrategy item
            "ignore_labels": ['O', 'X', 'S', "E"], // Out, Out, Start, End
        }
    }
    """

    def __init__(self, config: Dict):
        super(SeqLabPostProcessorConfig, self).__init__(config)

        self.use_crf = self.config['use_crf']
        self.word_ready = self.config['word_ready']
        self.aggregation_strategy = self.config['aggregation_strategy']
        self.ignore_labels = set(self.config['ignore_labels'])

        self.sentence = self.origin_input_map['sentence']
        self.offsets = self.origin_input_map['offsets']
        self.entities_info = self.origin_input_map['entities_info']
        self.uuid = self.origin_input_map['uuid']
        self.word_ids = self.origin_input_map['word_ids']
        self.special_tokens_mask = self.origin_input_map['special_tokens_mask']
        self.input_ids = self.origin_input_map['input_ids']
        self.label_ids = self.origin_input_map['label_ids']

        self.logits = self.input_map['logits']
        self.predict_seq_label = self.input_map['predict_seq_label']
        self._index = self.input_map['_index']

        if isinstance(self.config['meta'], str):
            meta = pkl.load(open(self.config['meta'], 'rb'))
        else:
            raise PermissionError("You must provide meta data(vocab & tokenizer) for ner postprocess.")

        vocab_trace_path = []
        vocab_trace_path_str = self.config['meta_data']['label_vocab']
        if vocab_trace_path_str and vocab_trace_path_str.strip()!='.':
            vocab_trace_path = vocab_trace_path_str.split('.')
        assert vocab_trace_path, "We need vocab and tokenizer all in meta, so you must provide the trace path from meta"
        self.label_vocab = meta
        for trace in vocab_trace_path:
            self.label_vocab = self.label_vocab[trace]
        self.label_vocab = Vocabulary.load(self.label_vocab)


        tokenizer_trace_path = []
        tokenizer_trace_path_str = self.config['meta_data']['tokenizer']
        if tokenizer_trace_path_str and tokenizer_trace_path_str.strip()!='.':
            tokenizer_trace_path = tokenizer_trace_path_str.split('.')
        assert tokenizer_trace_path, "We need vocab and tokenizer all in meta, so you must provide the trace path from meta"
        tokenizer_config = meta
        for trace in tokenizer_trace_path:
            tokenizer_config = tokenizer_config[trace]
        self.tokenizer = Tokenizer.from_str(tokenizer_config)

        self.save_path = self.config['save_path']
        self.save_root_path = self.config['save_root_path']
        self.start_save_epoch = self.config['start_save_epoch']
        self.start_save_step = self.config['start_save_step']

        self.post_check(self.config, used=[
            "meta",
            "use_crf",
            "word_ready",
            "meta_data",
            "input_map",
            "origin_input_map",
            "save_root_path",
            "save_path",
            "start_save_step",
            "start_save_epoch",
            "aggregation_strategy",
            "ignore_labels",
        ])


class AggregationStrategy(object):
    """docstring for AggregationStrategy"""
    NONE = "none"
    SIMPLE = "simple"
    FIRST = "first"
    AVERAGE = "average"
    MAX = "max"


@postprocessor_register('seq_lab')
class SeqLabPostProcessor(IPostProcessor):
    """docstring for DataSet"""
    def __init__(self, config:    SeqLabPostProcessorConfig):
        super(   SeqLabPostProcessor, self).__init__()
        self.config = config
        self.label_vocab = self.config.label_vocab
        self.tokenizer = self.config.tokenizer
        self.metric = torchmetrics.Accuracy()

    def do_predict(self, stage, list_batch_outputs, origin_data, rt_config)->List:
        """TODO: Docstring for do_predict.
        :stage: TODO
        :list_batch_outputs: the batch_output means the input
        :origin_data: the origin_data means the origin_input
        :rt_config: TODO
        :returns: TODO
        """
        predicts = []
        if self.config.use_crf:
            predicts = self.crf_predict(list_batch_outputs=list_batch_outputs, origin_data=origin_data)
        elif self.config.word_ready:
            predicts = self.word_predict(list_batch_outputs=list_batch_outputs, origin_data=origin_data)
        else:
            predicts = self.predict(list_batch_outputs=list_batch_outputs, origin_data=origin_data)
        return predicts

    def do_calc_metrics(self, predicts, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for do_calc_metrics.
        :returns: TODO

        """
        def _flat_entities_info(entities_info, text):
            """TODO: Docstring for _flat_entity_info.

            :entities_info: TODO
            :returns: TODO

            """
            info = {}
            for item in entities_info:
                label = item['labels'][0]
                if label not in info:
                    info[label] = []
                info[label].append(text[item['start']: item['end']].strip())
            return info

        all_predicts = []
        all_ground_truths = []
        for predict in predicts:
            text = predict['sentence']
            predict_ins = _flat_entities_info(predict['predict_entities_info'], text)
            ground_truth_ins = _flat_entities_info(predict['entities_info'], text)
            all_predicts.append(predict_ins)
            all_ground_truths.append(ground_truth_ins)

        precision, recall, f1 = self.calc_score(all_predicts, all_ground_truths)
        real_name = self.loss_name_map(stage)
        logger.info(f'{real_name}_precision: {precision*100}, {real_name}_recall: {recall*100}, {real_name}_f1: {f1*100}')
        return {f'{real_name}_precision': precision*100, f'{real_name}_recall': recall*100, f'{real_name}_f1': f1*100}

    def do_save(self, predicts, stage, rt_config={}, save_condition=False):
        """TODO: Docstring for do_save.

        :predicts: TODO
        :rt_config: TODO
        :condition: when the save condition is True, do save
        :returns: TODO
        """
        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get('total_steps', 0) - 1
            self.config.start_save_epoch = rt_config.get('total_epochs', 0) - 1
        if rt_config['current_step']>=self.config.start_save_step or rt_config['current_epoch']>=self.config.start_save_epoch:
            save_condition = True
        if save_condition:
            save_path = os.path.join(self.config.save_root_path, self.config.save_path.get(stage, ''))
            if not os.path.exists(save_path):
                os.makedirs(save_path, exist_ok=True)
            if "current_step" in rt_config:
                save_file = os.path.join(save_path, f"step_{str(rt_config['current_step'])}_predict.json")
            else:
                save_file = os.path.join(save_path, 'predict.json')
            logger.info(f"Save the {stage} predict data at {save_file}")
            json.dump(predicts, open(save_file, 'w'), indent=4)

    def calc_score(self, predict_list, ground_truth_list):
        """TODO: Docstring for calc_score.
        """
        category_tp = {}
        category_fp = {}
        category_fn = {}
        def _care_div(a, b):
            """TODO: Docstring for _care_div.
            """
            if b==0:
                return 0.0
            return a/b


        def calc_num(pred, ground_truth):
            """TODO: Docstring for calc_num.
            :returns: tp, fn, fp

            """
            num_p = len(pred)
            num_t = len(ground_truth)
            truth = 0
            for p in pred:
                if p in ground_truth:
                    truth += 1
            return truth, num_t-truth, num_p-truth

        for predict, ground_truth in zip(predict_list, ground_truth_list):
            keys = set(list(predict.keys())+list(ground_truth.keys()))
            for key in keys:
                tp, fn, fp = calc_num(predict.get(key, []), ground_truth.get(key, []))
                # logger.info(f"{key} tp num {tp}, fn num {fn}, fp num {fp}")
                category_tp[key] = category_tp.get(key, 0) + tp
                category_fn[key] = category_fn.get(key, 0) + fn
                category_fp[key] = category_fp.get(key, 0) + fp

        all_tp, all_fn, all_fp = 0, 0, 0
        for key in category_tp:
            tp, fn, fp = category_tp[key], category_fn[key], category_fp[key]
            all_tp += tp
            all_fn += fn
            all_fp += fp
            precision = _care_div(tp, tp+fp)
            recall = _care_div(tp, tp+fn)
            f1 = _care_div(2*precision*recall, precision+recall)
            logger.info(f"For entity 「{key}」, the precision={precision*100 :.2f}%, the recall={recall*100:.2f}%, f1={f1*100:.2f}%")
        precision = _care_div(all_tp,all_tp+all_fp)
        recall = _care_div(all_tp, all_tp+all_fn)
        f1 = _care_div(2*precision*recall, precision+recall)
        return precision, recall, f1

    def get_entity_info(self, sub_tokens_index, offset_mapping, word_ids, pre_label):
        """gather sub_tokens to get the start and end
        :returns: start, end info
        """
        if not sub_tokens_index or not pre_label:
            return None
        # TODO: use word_ids to combine the incomplete word
        start = offset_mapping[sub_tokens_index[0]][0]
        end = offset_mapping[sub_tokens_index[-1]][1]
        return {
            "start": start,
            "end": end,
            "labels": [pre_label]
        }

    def _process4predict(self, predict, index, origin_data):
        """TODO: Docstring for _process4predict.

        :predict: TODO
        :index: TODO
        :returns: TODO
        """
        one_ins = {}
        origin_ins = origin_data.iloc[int(index)]
        one_ins["sentence"] = origin_ins[self.config.sentence]
        one_ins["uuid"] = origin_ins[self.config.uuid]
        one_ins["entities_info"] = origin_ins[self.config.entities_info]

        word_ids = origin_ins[self.config.word_ids]
        rel_token_len = len(word_ids)
        offset_mapping = origin_ins[self.config.offsets][:rel_token_len]
        predict = list(predict[:rel_token_len])
        # predict = list(origin_ins['label_ids'][:rel_token_len])
        predict_entities_info = []
        pre_label = ''
        sub_tokens_index = []
        for i, label_id in enumerate(predict):
            if offset_mapping[i] == (0, 0): # added token like [CLS]/<s>/..
                continue
            label = self.config.label_vocab[label_id]
            if label in self.config.ignore_labels \
                or (label[0]=='B') \
                or (label.split('-')[-1] != pre_label):   # label == "O" or label=='B' or label.tail != previor_label
                entity_info = self.get_entity_info(sub_tokens_index, offset_mapping, word_ids, pre_label)
                if entity_info:
                    predict_entities_info.append(entity_info)
                pre_label = ''
                sub_tokens_index = []
            if label not in self.config.ignore_labels:
                assert len(label.split('-')) == 2
                pre_label = label.split('-')[-1]
                sub_tokens_index.append(i)
        entity_info = self.get_entity_info(sub_tokens_index, offset_mapping, word_ids, pre_label)
        if entity_info:
            predict_entities_info.append(entity_info)
        one_ins['predict_entities_info'] = predict_entities_info
        return one_ins

    def crf_predict(self, list_batch_outputs, origin_data):
        """TODO: Docstring for predict.

        :list_batch_outputs: TODO
        :origin_data: TODO
        :returns: TODO

        """
        if self.config.sentence not in origin_data:
            logger.error(f"{self.config.sentence} not in the origin data")
            raise PermissionError(f"{self.config.sentence} must be provided")
        if self.config.uuid not in origin_data:
            logger.error(f"{self.config.uuid} not in the origin data")
            raise PermissionError(f"{self.config.uuid} must be provided")
        if self.config.entities_info not in origin_data:
            logger.error(f"{self.config.entities_info} not in the origin data")
            raise PermissionError(f"{self.config.entities_info} must be provided")

        predicts = []
        for outputs in list_batch_outputs:
            batch_predict = outputs[self.config.predict_seq_label]
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config._index])
            outputs = []

            for predict, index in zip(batch_predict, indexes):
                one_ins = self._process4predict(predict, index, origin_data)
                predicts.append(one_ins)
        return predicts

    def word_predict(self, list_batch_outputs, origin_data):
        """TODO: Docstring for word_predict.

        :list_batch_outputs: TODO
        :origin_data: TODO
        :returns: TODO

        """
        if self.config.sentence not in origin_data:
            logger.error(f"{self.config.sentence} not in the origin data")
            raise PermissionError(f"{self.config.sentence} must be provided")
        if self.config.uuid not in origin_data:
            logger.error(f"{self.config.uuid} not in the origin data")
            raise PermissionError(f"{self.config.uuid} must be provided")
        if self.config.entities_info not in origin_data:
            logger.error(f"{self.config.entities_info} not in the origin data")
            raise PermissionError(f"{self.config.entities_info} must be provided")

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.logits].detach()
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config._index])

            outputs = []

            for logits, index in zip(batch_logits, indexes):
                origin_ins = origin_data.iloc[int(index)]
                word_ids = origin_ins[self.config.word_ids]

                rel_token_len = len(word_ids)
                logits = logits[:rel_token_len].cpu().numpy()

                predict = logits.argmax(-1)
                one_ins = self._process4predict(predict, index, origin_data)
                predicts.append(one_ins)
        return predicts

    def predict(self, list_batch_outputs, origin_data):
        """TODO: Docstring for predict.

        :list_batch_outputs: TODO
        :origin_data: TODO
        :returns: TODO

        """
        if self.config.sentence not in origin_data:
            logger.error(f"{self.config.sentence} not in the origin data")
            raise PermissionError(f"{self.config.sentence} must be provided")
        if self.config.uuid not in origin_data:
            logger.error(f"{self.config.uuid} not in the origin data")
            raise PermissionError(f"{self.config.uuid} must be provided")
        if self.config.entities_info not in origin_data:
            logger.error(f"{self.config.entities_info} not in the origin data")
            raise PermissionError(f"{self.config.entities_info} must be provided")

        predicts = []
        for outputs in list_batch_outputs:
            batch_logits = outputs[self.config.logits].detach()
            # batch_special_tokens_mask = outputs[self.config.special_tokens_mask]

            indexes = list(outputs[self.config._index])

            outputs = []

            for logits, index in zip(batch_logits, indexes):
                one_ins = {}
                origin_ins = origin_data.iloc[int(index)]

                input_ids = origin_ins[self.config.input_ids]
                one_ins["sentence"] = origin_ins[self.config.sentence]
                one_ins["uuid"] = origin_ins[self.config.uuid]
                one_ins["entities_info"] = origin_ins[self.config.entities_info]

                rel_token_len = len(input_ids)

                special_tokens_mask = np.array(origin_data.iloc[int(index)][self.config.special_tokens_mask][:rel_token_len])
                offset_mapping = origin_data.iloc[int(index)][self.config.offsets][:rel_token_len]

                logits = logits[:rel_token_len].cpu().numpy()

                entity_idx = logits.argmax(-1)
                labels = []
                for i, idx in enumerate(list(entity_idx)):
                    labels.append(self.config.label_vocab[idx])

                maxes = np.max(logits, axis=-1, keepdims=True)
                shifted_exp = np.exp(logits - maxes)
                scores = shifted_exp / shifted_exp.sum(axis=-1, keepdims=True)

                pre_entities = self.gather_pre_entities(
                    one_ins["sentence"], input_ids, scores, offset_mapping, special_tokens_mask)
                grouped_entities = self.aggregate(pre_entities, self.config.aggregation_strategy)
                # Filter anything that is in self.ignore_labels
                entities = [
                    entity
                    for entity in grouped_entities
                    if entity.get("entity", None) not in self.config.ignore_labels
                    and entity.get("entity_group", None) not in self.config.ignore_labels
                ]
                predict_entities_info = []
                for entity in entities:
                    one_predict = {}
                    one_predict['start'] = entity['start']
                    one_predict['end'] = entity['end']
                    one_predict['labels'] = [entity['entity_group']]
                    predict_entities_info.append(one_predict)
                one_ins['predict_entities_info'] = predict_entities_info
                predicts.append(one_ins)
        return predicts

    def aggregate(self, pre_entities: List[dict], aggregation_strategy: AggregationStrategy) -> List[dict]:
        if aggregation_strategy in {AggregationStrategy.NONE, AggregationStrategy.SIMPLE}:
            entities = []
            for pre_entity in pre_entities:
                entity_idx = pre_entity["scores"].argmax()
                score = pre_entity["scores"][entity_idx]
                entity = {
                    "entity": self.config.label_vocab[entity_idx],
                    "score": score,
                    "index": pre_entity["index"],
                    "word": pre_entity["word"],
                    "start": pre_entity["start"],
                    "end": pre_entity["end"],
                }
                entities.append(entity)
        else:
            entities = self.aggregate_words(pre_entities, aggregation_strategy)

        if aggregation_strategy == AggregationStrategy.NONE:
            return entities
        return self.group_entities(entities)

    def aggregate_word(self, entities: List[dict], aggregation_strategy: AggregationStrategy) -> dict:
        word = self.tokenizer.decode([self.tokenizer.token_to_id(entity['word']) for entity in entities])
        if aggregation_strategy == AggregationStrategy.FIRST:
            scores = entities[0]["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.config.label_vocab[idx]
        elif aggregation_strategy == AggregationStrategy.MAX:
            max_entity = max(entities, key=lambda entity: entity["scores"].max())
            scores = max_entity["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.config.label_vocab[idx]
        elif aggregation_strategy == AggregationStrategy.AVERAGE:
            scores = np.stack([entity["scores"] for entity in entities])
            average_scores = np.nanmean(scores, axis=0)
            entity_idx = average_scores.argmax()
            entity = self.config.label_vocab[entity_idx]
            score = average_scores[entity_idx]
        else:
            raise ValueError("Invalid aggregation_strategy")
        new_entity = {
            "entity": entity,
            "score": score,
            "word": word,
            "start": entities[0]["start"],
            "end": entities[-1]["end"],
        }
        return new_entity


    def group_sub_entities(self, entities: List[dict]) -> dict:
        """
        Group together the adjacent tokens with the same entity predicted.

        Args:
            entities (:obj:`dict`): The entities predicted by the pipeline.
        """
        # Get the first entity in the entity group
        entity = entities[0]["entity"].split("-")[-1]
        scores = np.nanmean([entity["score"] for entity in entities])
        tokens = [entity["word"] for entity in entities]
        entity_group = {
            "entity_group": entity,
            "score": np.mean(scores),
            "word": " ".join(tokens),
            "start": entities[0]["start"],
            "end": entities[-1]["end"],
        }
        return entity_group

    def get_tag(self, entity_name: str) -> Tuple[str, str]:
        if entity_name.startswith("B-"):
            bi = "B"
            tag = entity_name[2:]
        elif entity_name.startswith("I-"):
            bi = "I"
            tag = entity_name[2:]
        else:
            # It's not in B-, I- format
            # Default to I- for continuation.
            bi = "I"
            tag = entity_name
        return bi, tag

    def group_entities(self, entities: List[dict]) -> List[dict]:
        """
        Find and group together the adjacent tokens with the same entity predicted.

        Args:
            entities (:obj:`dict`): The entities predicted by the pipeline.
        """

        entity_groups = []
        entity_group_disagg = []

        for entity in entities:
            if not entity_group_disagg:
                entity_group_disagg.append(entity)
                continue

            # If the current entity is similar and adjacent to the previous entity,
            # append it to the disaggregated entity group
            # The split is meant to account for the "B" and "I" prefixes
            # Shouldn't merge if both entities are B-type
            bi, tag = self.get_tag(entity["entity"])
            last_bi, last_tag = self.get_tag(entity_group_disagg[-1]["entity"])

            if tag == last_tag and bi != "B":
                # Modify subword type to be previous_type
                entity_group_disagg.append(entity)
            else:
                # If the current entity is different from the previous entity
                # aggregate the disaggregated entity group
                entity_groups.append(self.group_sub_entities(entity_group_disagg))
                entity_group_disagg = [entity]
        if entity_group_disagg:
            # it's the last entity, add it to the entity groups
            entity_groups.append(self.group_sub_entities(entity_group_disagg))

        return entity_groups

    def aggregate_words(self, entities: List[dict], aggregation_strategy: AggregationStrategy) -> List[dict]:
        """
        Override tokens from a given word that disagree to force agreement on word boundaries.

        Example: micro|soft| com|pany| B-ENT I-NAME I-ENT I-ENT will be rewritten with first strategy as microsoft|
        company| B-ENT I-ENT
        """
        if aggregation_strategy in {
            AggregationStrategy.NONE,
            AggregationStrategy.SIMPLE,
        }:
            raise ValueError("NONE and SIMPLE strategies are invalid for word aggregation")

        word_entities = []
        word_group = None
        for entity in entities:
            if word_group is None:
                word_group = [entity]
            elif entity["is_subword"]:
                word_group.append(entity)
            else:
                word_entities.append(self.aggregate_word(word_group, aggregation_strategy))
                word_group = [entity]
        # Last item
        word_entities.append(self.aggregate_word(word_group, aggregation_strategy))
        return word_entities

    def gather_pre_entities(
        self,
        sentence: str,
        input_ids: np.ndarray,
        scores: np.ndarray,
        offset_mapping: Optional[List[Tuple[int, int]]],
        special_tokens_mask: np.ndarray,
    ) -> List[dict]:
        """Fuse various numpy arrays into dicts with all the information needed for aggregation"""
        pre_entities = []
        for idx, token_scores in enumerate(scores):
            # Filter special_tokens, they should only occur
            # at the sentence boundaries since we're not encoding pairs of
            # sentences so we don't have to keep track of those.
            if special_tokens_mask[idx]:
                continue

            word = self.tokenizer.id_to_token(int(input_ids[idx]))
            if offset_mapping is not None:
                start_ind, end_ind = offset_mapping[idx]
                word_ref = sentence[start_ind:end_ind]
                if getattr(self.tokenizer.model, "continuing_subword_prefix", None):
                    # This is a BPE, word aware tokenizer, there is a correct way
                    # to fuse tokens
                    is_subword = len(word) != len(word_ref)
                else:
                    # This is a fallback heuristic. This will fail most likely on any kind of text + punctuation mixtures that will be considered "words". Non word aware models cannot do better than this unfortunately.
                    is_subword = sentence[start_ind - 1 : start_ind] != " " if start_ind > 0 else False

            else:
                start_ind = None
                end_ind = None
                is_subword = False

            pre_entity = {
                "word": word,
                "scores": token_scores,
                "start": start_ind,
                "end": end_ind,
                "index": idx,
                "is_subword": is_subword,
            }
            pre_entities.append(pre_entity)
        return pre_entities




"""
Relabel the json data to bio
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('seq_lab_relabel')
class SeqLabRelabelConfig(BaseConfig):
    """docstring for SeqLabRelabelConfig
        {
            "_name": "seq_lab_relabel",
            "config": {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "input_map": {  // without necessery, don't change this
                        "word_ids": "word_ids",
                        "offsets": "offsets",
                        "entities_info": "entities_info",
                    },
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "output_map": {
                        "labels": "labels",
                    },
                    "start_label": "S",
                    "end_label": "E",
                }, //3
                "predict": "train",
                "online": "train",
            }
        }
    """
    def __init__(self, stage, config: Dict):

        super(SeqLabRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.offsets = self.config['input_map']['offsets']
        self.entities_info = self.config['input_map']['entities_info']
        self.start_label = self.config['start_label']
        self.end_label = self.config['end_label']
        self.output_labels = self.config['output_map']['labels']
        self.post_check(self.config, used=[
            "input_map",
            "data_set",
            "output_map",
            "start_label",
            "end_label"
        ])


@subprocessor_register('seq_lab_relabel')
class SeqLabRelabel(ISubProcessor):
    """docstring for SeqLabRelabel
    """

    def __init__(self, stage: str, config: SeqLabRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'seq_lab_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do seq_lab_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            data_set[self.config.output_labels] = data_set.parallel_apply(self.relabel, axis=1)

        return data

    def find_in_tuple(self, key, tuple_list, sub_word_ids, start, length, is_start=False):
        """TODO: Docstring for find.

        :key: TODO
        :tuple_list: TODO
        :start: TODO
        :returns: TODO
        """
        while start<length:
            if sub_word_ids[start] is None:
                start += 1
            elif key>=tuple_list[start][0] and key<tuple_list[start][1]:
                return start
            elif key<tuple_list[start][0]:
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins):
        """TODO: Docstring for relabel.
        :returns: TODO
        """
        pre_clean_entities_info = one_ins[self.config.entities_info]
        pre_clean_entities_info.sort(key=lambda x: x['start'])
        offsets = one_ins[self.config.offsets]
        sub_word_ids = one_ins[self.config.word_ids]
        if not sub_word_ids:
            logger.warning(f"entity_info: {pre_clean_entities_info}, offsets: {offsets} ")

        entities_info = []
        pre_end = -1
        pre_length = 0
        for entity_info in pre_clean_entities_info:
            assert len(entity_info['labels']) == 1, f"currently we just support one label for one entity"
            if entity_info['start']<pre_end:
                if entity_info['end'] - entity_info['start'] > pre_length:
                    entities_info.pop()
                else:
                    continue
            entities_info.append(entity_info)
            pre_end = entity_info['end']
            pre_length = entity_info['end'] - entity_info['start']

        cur_token_index = 0
        offset_length = len(offsets)
        sub_labels = []
        for entity_info in entities_info:
            start_token_index = self.find_in_tuple(entity_info['start'], offsets, sub_word_ids, cur_token_index, offset_length, is_start=True)
            if start_token_index == -1:
                logger.warning(f"cannot find the entity_info : {entity_info}, offsets: {offsets} ")
                continue
            for _ in range(start_token_index-cur_token_index):
                sub_labels.append('O')
            end_token_index = self.find_in_tuple(entity_info['end']-1, offsets, sub_word_ids, start_token_index, offset_length)
            assert end_token_index != -1, f"entity_info: {entity_info}, offsets: {offsets}"
            sub_labels.append("B-"+entity_info['labels'][0])
            for _ in range(end_token_index-start_token_index):
                sub_labels.append("I-"+entity_info['labels'][0])
            cur_token_index = end_token_index + 1
        assert cur_token_index<=offset_length
        for _ in range(offset_length-cur_token_index):
            sub_labels.append('O')

        if sub_word_ids[0] is None:
            sub_labels[0] = self.config.start_label

        if sub_word_ids[offset_length-1] is None:
            sub_labels[offset_length-1] = self.config.end_label

        if len(sub_labels)!= offset_length:
            logger.error(f"{len(sub_labels)} vs {offset_length}")
            for i in one_ins:
                logger.error(f"{i}")
            raise PermissionError

        return sub_labels



from typing import Dict
import torch.nn as nn
from dlk.utils.config import BaseConfig
from . import scheduler_register, scheduler_config_register, BaseScheduler
from torch.optim.lr_scheduler import LambdaLR
import torch.optim as optim


@scheduler_config_register("constant")
class ConstantScheduleConfig(BaseConfig):
    """
    {
        "config": {
            "last_epoch": -1
        },
        "_name": "constant",
    }
    """
    def __init__(self, config: Dict):
        super(ConstantScheduleConfig, self).__init__(config)
        config = config['config']
        self.last_epoch = config["last_epoch"]
        self.post_check(config, used=[
            "last_epoch",
        ])


@scheduler_register("constant")
class ConstantSchedule(BaseScheduler):
    def __init__(self, optimizer: optim.Optimizer, config: ConstantScheduleConfig):
        super(ConstantSchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self):
        """TODO: Docstring for get_scheduler.
        :returns: TODO
        """
        return LambdaLR(self.optimizer, lambda _: 1, last_epoch=self.config.last_epoch)



import hjson
import pickle as pkl
import os
import json
from typing import Union, Dict, Any
from dlk.utils.parser import BaseConfigParser, PostProcessorConfigParser
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.logger import Logger
import torch
from dlk.utils.vocab import Vocabulary
import torchmetrics

logger = Logger.get_logger()


@postprocessor_config_register('txt_reg')
class TxtRegPostProcessorConfig(IPostProcessorConfig):
    """docstring for TxtRegPostProcessorConfig
    config e.g.
    {
        "_name": "txt_reg",
        "config": {
            "input_map": {
                "logits": "logits",
                "values": "values",
                "_index": "_index",
            },
            "origin_input_map": {
                "sentence": "sentence",
                "sentence_a": "sentence_a", // for pair
                "sentence_b": "sentence_b",
                "uuid": "uuid"
            },
            "data_type": "single", //single or pair
            "save_root_path": ".",  //save data root dir
            "save_path": {
                "valid": "valid",  // relative dir for valid stage
                "test": "test",    // relative dir for test stage
            },
            "log_reg": false, // whether logistic regression
            "start_save_step": 0,  // -1 means the last
            "start_save_epoch": -1,
        }
    }
    """

    def __init__(self, config: Dict):
        super(TxtRegPostProcessorConfig, self).__init__(config)

        self.data_type = self.config['data_type']
        assert self.data_type in {'single', 'pair'}
        if self.data_type == 'pair':
            self.sentence_a = self.origin_input_map['sentence_a']
            self.sentence_b = self.origin_input_map['sentence_b']
        else:
            self.sentence = self.origin_input_map['sentence']
        self.uuid = self.origin_input_map['uuid']
        self.log_reg = self.config['log_reg']

        self.value = self.input_map['values']
        self.logits = self.input_map['logits']
        self._index = self.input_map['_index']
        self.save_path = self.config['save_path']
        self.save_root_path = self.config['save_root_path']
        self.start_save_epoch = self.config['start_save_epoch']
        self.start_save_step = self.config['start_save_step']
        self.post_check(self.config, used=[
            "input_map",
            "origin_input_map",
            "save_root_path",
            "save_path",
            "data_type",
            "start_save_step",
            "start_save_epoch",
            "log_reg",
        ])


@postprocessor_register('txt_reg')
class TxtRegPostProcessor(IPostProcessor):
    """docstring for TxtRegPostProcessor"""
    def __init__(self, config: TxtRegPostProcessorConfig):
        super(TxtRegPostProcessor, self).__init__()
        self.config = config

    def do_predict(self, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for do_predict.
        :stage: TODO
        :list_batch_outputs: the batch_output means the input
        :origin_data: the origin_data means the origin_input
        :rt_config: TODO
        :returns: TODO
        """
        results = []
        for outputs in list_batch_outputs:
            logits = outputs[self.config.logits].detach()
            if self.config.log_reg:
                logits = torch.sigmoid(logits)
            assert len(logits.shape) == 2
            # predict_indexes = list(torch.argmax(logits, 1))
            indexes = list(outputs[self.config._index])

            if self.config.value in outputs:
                values = outputs[self.config.value]
            else:
                values = [0.0] * len(indexes)
            for one_logits, index, value in zip(logits, indexes, values):
                one_ins = {}
                one_origin = origin_data.iloc[int(index)]
                if self.config.data_type == 'single':
                    sentence = one_origin[self.config.sentence]
                    one_ins['sentence'] = sentence
                else:
                    sentence_a = one_origin[self.config.sentence_a]
                    one_ins['sentence_a'] = sentence_a
                    sentence_b = one_origin[self.config.sentence_b]
                    one_ins['sentence_b'] = sentence_b
                    
                uuid = one_origin[self.config.uuid]
                one_ins['uuid'] = uuid
                one_ins['values'] = [float(value)]
                one_ins['predict_values'] = [float(one_logits)]
                results.append(one_ins)
        return results

    def do_calc_metrics(self, predicts, stage, list_batch_outputs, origin_data, rt_config):
        """TODO: Docstring for do_calc_metrics.
        :returns: TODO

        """
        return {}

    def do_save(self, predicts, stage, list_batch_outputs, origin_data, rt_config={}, save_condition=False):
        """TODO: Docstring for do_save.

        :predicts: TODO
        :rt_config: TODO
        :condition: when the save condition is True, do save
        :returns: TODO
        """
        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get('total_steps', 0) - 1
            self.config.start_save_epoch = rt_config.get('total_epochs', 0) - 1
        if rt_config['current_step']>=self.config.start_save_step or rt_config['current_epoch']>=self.config.start_save_epoch:
            save_condition = True
        if save_condition:
            save_path = os.path.join(self.config.save_root_path, self.config.save_path.get(stage, ''))
            if not os.path.exists(save_path):
                os.makedirs(save_path, exist_ok=True)
            if "current_step" in rt_config:
                save_file = os.path.join(save_path, f"step_{str(rt_config['current_step'])}_predict.json")
            else:
                save_file = os.path.join(save_path, 'predict.json')
            logger.info(f"Save the {stage} predict data at {save_file}")
            json.dump(predicts, open(save_file, 'w'), indent=4)



"""
Relabel the json data to bio.
And gather the firstpiece subword index to deliver $gather_index; for get the firstpiece index we must use the $word_ids(generator by tokenizer)
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger

logger = Logger.get_logger()


@subprocessor_config_register('seq_lab_firstpiece_relabel')
class SeqLabFirstPieceRelabelConfig(BaseConfig):
    """docstring for SeqLabFirstPieceRelabelConfig
        {
            "_name": "seq_lab_firstpiece_relabel",
            "config": {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "input_map": {  // without necessery, don't change this
                        "word_ids": "word_ids",
                        "offsets": "offsets",
                        "entities_info": "entities_info",
                    },
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test', 'predict'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "output_map": {
                        "labels": "labels",
                        "gather_index": "gather_index",
                        "word_word_ids": "word_ids",
                        "word_offsets": "offsets",
                    },
                    "start_label": "S",
                    "end_label": "E",
                }, //3
                "predict": "train",
                "online": "train",
            }
        }
    """
    def __init__(self, stage, config: Dict):

        super(SeqLabFirstPieceRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.word_word_ids = self.config['output_map']['word_word_ids']
        self.word_offsets = self.config['output_map']['word_offsets']
        self.offsets = self.config['input_map']['offsets']
        self.entities_info = self.config['input_map']['entities_info']
        self.start_label = self.config['start_label']
        self.end_label = self.config['end_label']
        self.gather_index = self.config['output_map']['gather_index']
        self.output_labels = self.config['output_map']['labels']
        self.post_check(self.config, used=[
            "input_map",
            "data_set",
            "output_map",
            "start_label",
            "end_label",
        ])


@subprocessor_register('seq_lab_firstpiece_relabel')
class SeqLabFirstPieceRelabel(ISubProcessor):
    """docstring for SeqLabFirstPieceRelabel
    """

    def __init__(self, stage: str, config: SeqLabFirstPieceRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'seq_lab_firstpiece_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do seq_lab_firstpiece_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            data_set[[self.config.output_labels,
                self.config.gather_index,
                self.config.word_word_ids,
                self.config.word_offsets]] = data_set.parallel_apply(self.relabel, axis=1, result_type="expand")
        return data

    def find_in_tuple(self, key, tuple_list, sub_word_ids, start, length, is_start=False):
        """TODO: Docstring for find.

        :key: TODO
        :tuple_list: TODO
        :start: TODO
        :returns: TODO
        """
        while start<length:
            if sub_word_ids[start] is None:
                start += 1
            elif key>=tuple_list[start][0] and key<tuple_list[start][1]:
                return start
            elif key<tuple_list[start][0]:
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins):
        """TODO: Docstring for relabel.
        :returns: TODO
        """
        pre_clean_entities_info = one_ins[self.config.entities_info]
        pre_clean_entities_info.sort(key=lambda x: x['start'])
        offsets = one_ins[self.config.offsets]
        sub_word_ids = one_ins[self.config.word_ids]
        if not sub_word_ids:
            logger.warning(f"entity_info: {pre_clean_entities_info}, offsets: {offsets} ")

        gather_index = []
        pre_word_id = -1
        word_offset = []
        word_offsets = []
        word_ids = []
        for i, (token_offset, word_id) in enumerate(zip(offsets, sub_word_ids)):
            if word_id != pre_word_id:
                gather_index.append(i)
                word_ids.append(word_id)
                if word_offset:
                    word_offsets.append(word_offset)
                word_offset = list(token_offset)
                pre_word_id = word_id
            else:
                assert word_offset
                word_offset[1] = token_offset[1]
        if word_offset:
            word_offsets.append(word_offset)

        entities_info = []
        pre_end = -1
        pre_length = 0
        for entity_info in pre_clean_entities_info:
            assert len(entity_info['labels']) == 1, f"currently we just support one label for one entity"
            if entity_info['start']<pre_end:
                if entity_info['end'] - entity_info['start'] > pre_length:
                    entities_info.pop()
                else:
                    continue
            entities_info.append(entity_info)
            pre_end = entity_info['end']
            pre_length = entity_info['end'] - entity_info['start']

        cur_token_index = 0
        offset_length = len(word_offsets)
        sub_labels = []
        for entity_info in entities_info:
            start_token_index = self.find_in_tuple(entity_info['start'], word_offsets, word_ids, cur_token_index, offset_length, is_start=True)
            if start_token_index == -1:
                logger.warning(f"cannot find the entity_info : {entity_info}, word_offsets: {word_offsets} ")
                continue
            for _ in range(start_token_index-cur_token_index):
                sub_labels.append('O')
            end_token_index = self.find_in_tuple(entity_info['end']-1, word_offsets, word_ids, start_token_index, offset_length)
            assert end_token_index != -1, f"entity_info: {entity_info}, word_offsets: {word_offsets}"
            sub_labels.append("B-"+entity_info['labels'][0])
            for _ in range(end_token_index-start_token_index):
                sub_labels.append("I-"+entity_info['labels'][0])
            cur_token_index = end_token_index + 1
        assert cur_token_index<=offset_length
        for _ in range(offset_length-cur_token_index):
            sub_labels.append('O')

        if word_ids[0] is None:
            sub_labels[0] = self.config.start_label

        if word_ids[offset_length-1] is None:
            sub_labels[offset_length-1] = self.config.end_label

        if len(sub_labels)!= len(gather_index):
            logger.error(f"{len(sub_labels)} vs {len(gather_index)}")
            for i in one_ins:
                logger.error(f"{i}")
            raise PermissionError

        return sub_labels, gather_index, word_ids, word_offsets



"""
Loader the data from dict and generator DataFrame
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
import pandas as pd
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('seq_lab_loader')
class SeqLabLoaderConfig(BaseConfig):
    """docstring for SeqLabLoaderConfig
        {
            "_name": "seq_lab_loader",
            "config": {
                "train":{ //train、predict、online stage config,  using '&' split all stages
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test', 'predict'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "input_map": {   // without necessery don't change this
                        "sentence": "sentence",
                        "uuid": "uuid",
                        "entities_info": "entities_info",
                    },
                    "output_map": {   // without necessery don't change this
                        "sentence": "sentence",
                        "uuid": "uuid",
                        "entities_info": "entities_info",
                    },
                }, //3
                "predict": "train",
                "online": "train",
            }
        }

        NOTE: the seq_labformat input is
        {
            "uuid": '**-**-**-**'
            "sentence": "I have an apple",
            "labels": [
                        {
                            "end": 15,
                            "start": 10,
                            "labels": [
                                "Fruit"
                            ]
                        },
                        ...,
                    ]
                },
            ],
        }
    """
    def __init__(self, stage, config: Dict):

        super(SeqLabLoaderConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.output_map = self.config.get('output_map', {})
        self.input_map = self.config.get('input_map', {})
        self.post_check(self.config, used=[
            "data_set",
            "input_map",
            "output_map",
        ])


@subprocessor_register('seq_lab_loader')
class SeqLabLoader(ISubProcessor):
    """docstring for SeqLabLoader
    """

    def __init__(self, stage: str, config: SeqLabLoaderConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'seq_lab_loader' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        '''
            data: {
                "train": list of json format train data
            }
        '''

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do seq_lab_loader on it.')
                continue
            data_set = data['data'][data_set_name]

            sentences = []
            uuids = []
            entities_infos = []

            for one_ins in data_set:
                try:
                    sentences.append(one_ins[self.config.input_map['sentence']])
                    uuids.append(one_ins[self.config.input_map['uuid']])
                    entities_infos.append(one_ins[self.config.input_map['entities_info']])
                except:
                    raise PermissionError(f"You must provide the data as requests, we need 'sentence', 'uuid' and 'entities_info', or you can provide the input_map to map the origin data to this format")
            data_df = pd.DataFrame(data= {
                self.config.output_map["sentence"]: sentences,
                self.config.output_map["uuid"]: uuids,
                self.config.output_map["entities_info"]: entities_infos,
            })
            data['data'][data_set_name] = data_df

        return data



"""
gather all tokens from the 'gather_columns' and deliver a vocab named 'token_vocab'
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('token_gather')
class TokenGatherConfig(BaseConfig):
    """Config eg.
        {
            "_name": "token_gather",
            "config": {
                "train": { // only train stage using
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ["train", "valid", 'test']
                    },
                    "gather_columns": "*@*", //List of columns. Every cell must be sigle token or list of tokens or set of tokens
                    "deliver": "*@*", // output Vocabulary object (the Vocabulary of labels) name.
                    "ignore": "", // ignore the token, the id of this token will be -1
                    "update": null, // null or another Vocabulary object to update
                    "unk": "[UNK]",
                    "pad": "[PAD]",
                    "min_freq": 1,
                    "most_common": -1, //-1 for all
                }
            }
        }
    """

    def __init__(self, stage: str, config: Dict):
        super(TokenGatherConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.ignore = self.config['ignore']
        self.gather_columns = self.config["gather_columns"]
        self.deliver = self.config["deliver"]
        if self.data_set and (not self.deliver):
            raise ValueError("The 'deliver' value must not be null.")
        self.update = self.config['update']
        self.unk = self.config['unk']
        self.pad = self.config['pad']
        self.min_freq = self.config['min_freq']
        self.most_common = self.config['most_common']
        self.post_check(self.config, used=[
            "data_set",
            "gather_columns",
            "deliver",
            "ignore",
            "update",
            "unk",
            "pad",
            "min_freq",
            "most_common",
        ])

@subprocessor_register('token_gather')
class TokenGather(ISubProcessor):
    """
    """
    def __init__(self, stage: str, config: TokenGatherConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'token_gather' at stage {self.stage}")
            return
        self.update = config.update

    def process(self, data: Dict)->Dict:
        if not self.data_set:
            return data
        if self.update:
            self.vocab = data[self.update]
        else:
            self.vocab = Vocabulary(do_strip=True, unknown=self.config.unk, ignore=self.config.ignore, pad=self.config.pad)
        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip gather tokens from it.')
                continue
            data_set = data['data'][data_set_name]
            for column in self.config.gather_columns:
                self.vocab.auto_update(data_set[column])
        self.vocab.filter_rare(self.config.min_freq, self.config.most_common)
        logger.info(f"The Vocab Num is {self.vocab.word_num}")
        data[self.config.deliver] = self.vocab.__dict__
        return data



"""processors"""

import importlib
import os
from typing import Callable, Dict, Type
from dlk.utils.register import Register
import abc

class IProcessor(metaclass=abc.ABCMeta):
    """docstring for IProcessor"""


    @abc.abstractmethod
    def process(self, data: Dict)->Dict:
        """TODO: Docstring for process.

        :arg1: TODO
        :returns: TODO

        """
        raise NotImplementedError


processor_config_register = Register('Processor config register')
processor_register = Register("Processor register")


def import_processors(processors_dir, namespace):
    for file in os.listdir(processors_dir):
        path = os.path.join(processors_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            # and not (file.endswith("subprocessors") and os.path.isdir(path))
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            processor_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + processor_name)


# automatically import any Python files in the models directory
processors_dir = os.path.dirname(__file__)
import_processors(processors_dir, "dlk.data.processors")



"""
    This file convert the conll03 format bio sequence labeling data to json format
    result is format as
    [{
        "uuid": '**-**-**-**'
        "sentence": "Mie Merah - Buah Bit - Noodles only - 1 box isi 4 - AM MA Kitchen - Red healthy noodle",
        "entities_info": [
                    {
                        "end": 65,
                        "start": 52,
                        "labels": [
                            "Brand"
                        ]
                    },
                    {
                        "end": 36,
                        "start": 22,
                        "labels": [
                            "Product"
                        ]
                    },
                    {
                        "end": 9,
                        "start": 0,
                        "labels": [
                            "Product"
                        ]
                    },
                    {
                        "end": 49,
                        "start": 38,
                        "labels": [
                            "Specification"
                        ]
                    },
                    {
                        "end": 86,
                        "start": 68,
                        "labels": [
                            "Health Benefits"
                        ]
                    }
                ]
            },
        ],
    },
    ]
"""
from tqdm import tqdm
import uuid
import json
import os


class Conll03Reader:
    def read(self, data_path):
        data_parts = ['train', 'valid', 'test']
        extension = '.txt'
        dataset = {}
        for data_part in tqdm(data_parts):
            file_path = os.path.join(data_path, data_part+extension)
            dataset[data_part] = self.read_file(str(file_path))

        format_dataset = {}
        for key, data in dataset.items():
            format_dataset[key] = self.convert(data)
        return format_dataset

    def convert(self, data):
        """TODO: Docstring for convert.
        :data: TODO
        :returns: TODO
        """
        format_data = []
        for line in data:
            tokens, labels = line[0], line[1]
            text = ""
            cur_label = ''
            start = -1
            entities_info = []
            entity_info = {}
            for token, label in zip(tokens, labels):
                assert label[0] in ['B', 'I', "O"]
                if label[0] == 'B':
                    if text:
                        cur_start = len(text) + 1 # will add space begin current token
                        text = " ".join([text, token])
                    else:
                        cur_start = 0
                        text = token
                    if cur_label:
                        entity_info['start'] = start
                        entity_info['end'] = len(text)
                        entity_info["labels"] = [cur_label]
                        entities_info.append(entity_info)
                    start = cur_start
                    cur_label = label.split('-')[-1]
                    entity_info = {}
                elif label[0] == 'O':
                    if cur_label:
                        entity_info['start'] = start
                        entity_info['end'] = len(text)
                        entity_info["labels"] = [cur_label]
                        entities_info.append(entity_info)
                    entity_info = {}
                    cur_label = ''
                    start = -1
                    if text:
                        text = " ".join([text, token])
                    else:
                        text = token
                else:
                    if text:
                        text = " ".join([text, token])
                    else:
                        text = token
            if cur_label:
                entity_info['start'] = start
                entity_info['end'] = len(text)
                entity_info['labels'] = [cur_label]
                entities_info.append(entity_info)
            for entity in entities_info:
                assert len(text[entity['start']: entity['end']].strip()) == entity['end'] - entity['start'], f"{entity}, {len(text[entity['start']: entity['end']].strip())},{entity['end'] - entity['start']},{text}"

            format_data.append({'uuid': str(uuid.uuid1()),  "sentence": text, "entities_info": entities_info})
        return format_data

    def read_file(self, file_path):
        samples = []
        tokens = []
        tags = []
        with open(file_path,'r', encoding='utf-8') as fb:
            for line in fb:
                line = line.strip('\n')

                if line == '-DOCSTART- -X- -X- O':
                    pass
                elif line =='':
                    if len(tokens) != 0:
                        samples.append((tokens, tags))
                        tokens = []
                        tags = []
                else:
                    contents = line.split(' ')
                    tokens.append(contents[0])
                    tags.append(contents[-1])
        return samples

if __name__ == "__main__":
    ds_rd = Conll03Reader()
    base_dir = "./data"
    data = ds_rd.read(base_dir)
    for key in data:
        json.dump(data[key], open(os.path.join(base_dir, key+'.json'), 'w'), indent=4)



from tokenizers.processors import TemplateProcessing
from tokenizers.normalizers import Lowercase, NFD, NFC, StripAccents, Strip
from tokenizers.pre_tokenizers import WhitespaceSplit, ByteLevel


class TokenizerPostprocessorFactory(object):
    """docstring for TokenizerPostprocessorFactory"""

    @property
    def bert(self):
        """TODO: Docstring for bert.

        :arg1: TODO
        :returns: TODO

        """
        def _wrap():
            """TODO: Docstring for _wrap.
            :returns: TODO

            """
            return TemplateProcessing(
                single="[CLS] $A [SEP]",
                pair="[CLS] $A [SEP] $B:1 [SEP]:1",
                special_tokens=[
                    ("[CLS]", 1),
                    ("[SEP]", 2),
                ],
            )
        return _wrap


    def get(self, name):
        """TODO: Docstring for get.
        """
        return self.__getattribute__(name)



class PreTokenizerFactory(object):
    """docstring for PreTokenizerFactory"""

    @property
    def bytelevel(self):
        """TODO: Docstring for bytelevel.

        :arg1: TODO
        :returns: TODO

        """
        return ByteLevel

    @property
    def whitespace(self):
        """TODO: Docstring for bytelevel.

        :arg1: TODO
        :returns: TODO

        """
        return WhitespaceSplit

    def get(self, name):
        """TODO: Docstring for get.

        :name: TODO
        :returns: TODO

        """
        return self.__getattribute__(name)


class TokenizerNormalizerFactory(object):
    """docstring for TokenizerNormalizerFactory"""

    @property
    def lowercase(self):
        """TODO: Docstring for lowercase.

        :arg1: TODO
        :returns: TODO

        """
        return Lowercase

    @property
    def nfd(self):
        """TODO: Docstring for nfd.

        :arg1: TODO
        :returns: TODO

        """
        return NFD

    @property
    def nfc(self):
        """TODO: Docstring for nfc.
        """
        return NFC

    @property
    def strip_accents(self):
        """
        """
        return StripAccents

    @property
    def strip(self):
        """
        """
        return Strip

    def get(self, name):
        """TODO: Docstring for get.
        """
        return self.__getattribute__(name)



from typing import Dict
from dlk.utils.config import BaseConfig
from . import scheduler_register, scheduler_config_register, BaseScheduler
from torch.optim.lr_scheduler import LambdaLR
from dlk.utils.logger import Logger
import torch.optim as optim
logger = Logger.get_logger()


@scheduler_config_register("linear_warmup")
class LinearWarmupScheduleConfig(BaseConfig):
    """
    {
        "config": {
            "last_epoch": -1,
            "num_warmup_steps": 0,
            "num_training_steps": -1,
        },
        "_name": "linear_warmup",
    }
    """
    def __init__(self, config: Dict):
        super(LinearWarmupScheduleConfig, self).__init__(config)
        config = config['config']
        self.last_epoch = config["last_epoch"]
        self.num_warmup_steps = config["num_warmup_steps"]
        self.num_training_steps = config["num_training_steps"]
        self.post_check(config, used=[
            "last_epoch",
            "num_warmup_steps",
            "num_training_steps",
        ])


@scheduler_register("linear_warmup")
class LinearWarmupSchedule(BaseScheduler):
    def __init__(self, optimizer: optim.Optimizer, config: LinearWarmupScheduleConfig):
        super(LinearWarmupSchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self):
        """TODO: Docstring for get_scheduler.
        :returns: TODO
        """
        num_training_steps = self.config.num_training_steps
        num_warmup_steps = self.config.num_warmup_steps
        if num_warmup_steps >0 and num_warmup_steps < 1:
            num_warmup_steps = int(num_warmup_steps * num_training_steps)
        last_epoch = self.config.last_epoch
        logger.warning(f"The calculated Total Traning Num is {num_training_steps}, the Num Warmup Steps is {num_warmup_steps}. Please check it carefully.")

        def lr_lambda(current_step: int):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1, num_warmup_steps))
            return max(
                0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
            )

        return LambdaLR(self.optimizer, lr_lambda, last_epoch)



from typing import Dict
import torch.nn as nn
from . import loss_register, loss_config_register
from packaging import version
import torch.nn as nn
from dlk.utils.logger import Logger
from dlk.core.base_module import BaseModuleConfig
import torch

logger = Logger.get_logger()

@loss_config_register("mse")
class MSELossConfig(BaseModuleConfig):
    """docstring for MSELossConfig
    {
        "config": {
            "pred_truth_pair": [], # len(.) == 2, the 1st is the pred_name, 2nd is truth_name in __call__ inputs
            "schedule": [1],
            "masked_select": null, // if provide, only select the masked(=1) data
            "scale": [1], # scale the loss for every schedule stage
            // "schdeule": [0.3, 1.0], # can be a list or str
            // "scale": "[0.5, 1]",
        },
        "_name": "mse",
    }
    """
    def __init__(self, config: Dict):
        super(MSELossConfig, self).__init__(config)
        config = config['config']

        self.scale = config['scale']
        self.schedule = config['schedule']

        if isinstance(self.scale, str):
            self.scale = eval(self.scale)
        if isinstance(self.schedule, str):
            self.schedule = eval(self.schedule)

        if not isinstance(self.scale, list):
            assert isinstance(float(self.scale), float)
            self.scale = [self.scale]
        if not isinstance(self.schedule, list):
            assert isinstance(float(self.schedule), float)
            self.schedule = [self.schedule]
        assert len(self.schedule) == len(self.scale)
        assert self.schedule[-1] - 1 < 0.00001

        self.pred_truth_pair = config['pred_truth_pair']
        if not self.pred_truth_pair:
            raise PermissionError(f"You must provide the pred_truth_pair for loss.")
        self.masked_select = config['masked_select']
        self.post_check(config, used=[
            "pred_truth_pair",
            "masked_select",
            "schedule",
            "scale",
        ])


@loss_register("mse")
class MSELoss(object):
    def __init__(self, config: MSELossConfig):
        super(MSELoss, self).__init__()
        self.config = config
        self.mse = nn.MSELoss(reduction='sum') # we will use masked_select, and apply reduction=batchmean (sum/batch_size)

    def update_config(self, rt_config):
        """TODO: Docstring for update_config.
        :rt_config: TODO
         {
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }
        :returns: TODO

        """
        self.current_stage = 0
        self.config.schedule = [rt_config['total_steps']*i for i in self.config.schedule]

    def calc(self, result, inputs, rt_config):
        """TODO: Docstring for get_loss.
        :returns: TODO
        """
        if rt_config['current_step']>self.config.schedule[self.current_stage]:
            self.current_stage += 1
        scale = self.config.scale[self.current_stage]
        pred_name, truth_name = self.config.pred_truth_pair
        pred = result[pred_name]
        target = inputs[truth_name]
        batch_size = target.shape[0]
        if self.config.masked_select is not None:
            pred = torch.masked_select(pred, inputs[self.config.masked_select])
            target = torch.masked_select(target, inputs[self.config.masked_select])
        loss = self.mse(pred, target) * scale / batch_size # batch mean
        return loss

    def __call__(self, result, inputs, rt_config):
        """TODO: Docstring for __call__.
        :returns: TODO
         rt_config={
             "current_step": self.global_step,
             "current_epoch": self.current_epoch,
             "total_steps": self.num_training_steps,
             "total_epochs": self.num_training_epochs
         }),
        """
        return self.calc(result, inputs, rt_config)



import logging
import os
import colorlog


class Logger(object):
    """docstring for logger"""
    global_logger = None
    global_log_file = None
    global_file_handler = None
    global_log_file = None
    color_config = {
        'DEBUG': 'white',  # cyan white
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'bold_red',
    }
    level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    def __init__(self, log_file: str='', base_dir: str='logs', log_level: str='debug', log_name='dlk'):
        super(Logger, self).__init__()
        self.log_file = log_file
        self.base_dir = base_dir
        if self.base_dir and not os.path.isdir(self.base_dir):
            os.mkdir(self.base_dir)

        self.init_global_logger(log_level=log_level, log_name=log_name)
        self.init_file_logger(log_file, base_dir, log_level=log_level)

    @staticmethod
    def get_logger()->logging.Logger:
        """TODO: Docstring for get_logger.
        :returns: TODO

        """
        if Logger.global_logger is None:
            Logger.init_global_logger()
            Logger.global_logger.warning("You didn't init the logger, so we use the default logger setting to output to terminal, you can always set the file logger by yourself.")
        return Logger.global_logger

    @staticmethod
    def init_file_logger(log_file, base_dir='logs', log_level: str='debug'):
        """TODO: Docstring for init_file_logger.
        :log_file: TODO
        :base_dir: TODO
        :log_level: TODO
        :returns: TODO
        """
        if log_file:
            log_file = os.path.join(base_dir, log_file)
        if log_file and log_file != Logger.global_log_file:
            if Logger.global_file_handler is not None:
                Logger.global_logger.removeHandler(Logger.global_file_handler)
            file_handler = logging.FileHandler(filename=log_file, mode='a', encoding='utf8')
            file_handler.setLevel(Logger.level_map[log_level])

            file_formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                                                      datefmt='%m/%d/%Y %H:%M:%S')
            file_handler.setFormatter(file_formatter)
            Logger.global_file_handler = file_handler
            Logger.global_logger.addHandler(file_handler)

    @staticmethod
    def init_global_logger(log_level: str='debug', log_name: str='dlk'):
        """TODO: Docstring for init_global_logger.
        :returns: TODO
        """
        if Logger.global_logger is None:
            Logger.global_logger = logging.getLogger(log_name)
            Logger.global_logger.setLevel(Logger.level_map[log_level])
            console_handler = logging.StreamHandler()
            console_handler.setLevel(Logger.level_map[log_level])
            console_formatter = colorlog.ColoredFormatter(
                fmt='%(log_color)s%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                datefmt='%m/%d/%Y %H:%M:%S',
                log_colors=Logger.color_config)
            console_handler.setFormatter(console_formatter)
            Logger.global_logger.addHandler(console_handler)



from dlk.utils.logger import Logger
import pytorch_lightning as pl
from dlk.train import Train
import json

logger = Logger('log.txt')
pl.seed_everything(88)

# trainer = Train('./examples/sequence_labeling/benchmark/pretrained/first_piece_lstm_crf_main.hjson')
trainer = Train('./distil_bert/main.hjson')
# print(json.dumps(trainer.configs, indent=4))

trainer.run()



"""callbacks"""

import importlib
import os
from typing import Dict, Any
from dlk.utils.register import Register

callback_config_register = Register("Callback config register")
callback_register = Register("Callback register")


def import_callbacks(callbacks_dir, namespace):
    for file in os.listdir(callbacks_dir):
        path = os.path.join(callbacks_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            callback_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + callback_name)


# automatically import any Python files in the models directory
callbacks_dir = os.path.dirname(__file__)
import_callbacks(callbacks_dir, "dlk.core.callbacks")



# import pickle as pkl

# train = pkl.load(open('./test_cls/meta.pkl', 'rb'))
# print(train)

# data = pkl.load(open('./test_cls/processed_data.pkl', 'rb'))
# print(data)



import pandas as pd
# from dlk.utils.logger import Logger
# setting_logger('process.log')
# logger = 
from dlk.utils.parser import config_parser_register
import copy
import json
import hjson
from dlk.process import Processor

# train_data = json.load(open('./local_data/tasks_data/title/title_train.json', 'r'))
# valid_data = json.load(open('./local_data/tasks_data/title/title_valid.json', 'r'))
# test_data = json.load(open('./local_data/tasks_data/title/title_test.json', 'r'))


# NER benchmark data
# train_data = json.load(open('./local_data/tasks_data/ner_benchmark/train.json', 'r'))
# valid_data = json.load(open('./local_data/tasks_data/ner_benchmark/valid.json', 'r'))
# test_data = json.load(open('./local_data/tasks_data/ner_benchmark/test.json', 'r'))
# train_data = json.load(open('./examples/sequence_labeling/conll2003/data/train.json', 'r'))
# valid_data = json.load(open('./examples/sequence_labeling/conll2003/data/valid.json', 'r'))
# test_data = json.load(open('./examples/sequence_labeling/conll2003/data/test.json', 'r'))
# data = {"train": train_data, "valid": valid_data, "test": test_data}
# data = {"test": test_data}

# inp = {"data": data}
# config = config_parser_register.get("processor")(hjson.load(open("./examples/sequence_labeling/conll2003/norm_lstm_crf/prepro.hjson"),object_pairs_hook=dict)).parser_with_check()[0]
processor = Processor('./examples/sequence_labeling/conll2003/norm_char_lstm_crf/prepro.hjson')
processor.fit({})



"""optimizers"""
import importlib
import os
from typing import Callable, Dict, Tuple, Any
from dlk.utils.register import Register
import torch.optim as optim
from dlk.utils.logger import Logger
import re


logger = Logger.get_logger()
optimizer_config_register = Register("Optimizer config register.")
optimizer_register = Register("Optimizer register.")


class BaseOptimizer(object):

    def get_optimizer(self):
        """TODO: Docstring for get_optimizer.
        :returns: TODO

        """
        raise NotADirectoryError

    def init_optimizer(self, optimizer, model, config):
        """TODO: Docstring for get_optimizer.
        :returns: TODO
        """
        optimizer_special_groups = config.pop('optimizer_special_groups', {})
        params = []
        all_named_parameters = list(model.named_parameters())
        total_all_named_parameters = len(all_named_parameters)
        logger.info(f"All Named Params Num is {len(all_named_parameters)}")
        has_grouped_params = set()
        for special_group_name in optimizer_special_groups.get('order', []):

            group_config = optimizer_special_groups[special_group_name]['config']
            group_patterns = optimizer_special_groups[special_group_name]['pattern']
            # convert to regex
            combine_patterns = []
            for pattern in group_patterns:
                combine_patterns.append(f"({pattern})")
            cc_patterns = re.compile("|".join(combine_patterns))
            group_param = []
            for n, p  in all_named_parameters:
                # logger.info(f"Param name {n}")
                if n in has_grouped_params:
                    continue
                if cc_patterns.search(n): # use regex
                    has_grouped_params.add(n)
                    group_param.append(p)
            group_config['params'] = group_param
            group_config['name'] = special_group_name
            params.append(group_config)

        reserve_params = [p for n, p in all_named_parameters if not n in has_grouped_params]
        params.append({"params": reserve_params, "name": config.pop('name')})
        logger.info(f"Param Group Nums {len(params)}")
        total_param = 0
        for group in params:
            total_param = total_param + len(group['params'])
        assert total_param == total_all_named_parameters

        return optimizer(params=params, **config)

    def __call__(self):
        """TODO: Docstring for __call__.
        :returns: TODO

        """
        return self.get_optimizer()


def import_optimizers(optimizers_dir, namespace):
    for file in os.listdir(optimizers_dir):
        path = os.path.join(optimizers_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            optimizer_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + optimizer_name)


# automatically import any Python files in the optimizers directory
optimizers_dir = os.path.dirname(__file__)
import_optimizers(optimizers_dir, "dlk.core.optimizers")



import torch.nn as nn
import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import encoder_register, encoder_config_register
from dlk.core.modules import module_config_register, module_register
from dlk.utils.logger import Logger
logger = Logger.get_logger()

@encoder_config_register("lstm")
class LSTMConfig(BaseModuleConfig):
    """docstring for LSTMConfig
    {
        module: {
            _base: "lstm",
        },
        config: {
            input_map: {},
            output_map: {},
            input_size: *@*,
            output_size: "*@*",
            num_layers: 1,
            dropout: "*@*", // dropout between layers
        },
        _link: {
            config.input_size: [module.config.input_size],
            config.output_size: [module.config.output_size],
            config.dropout: [module.config.dropout],
        },
        _name: "lstm",
    }
    """

    def __init__(self, config: Dict):
        super(LSTMConfig, self).__init__(config)
        self.lstm_config = config["module"]
        assert self.lstm_config['_name'] == "lstm"
        self.post_check(config['config'], used=[
            "input_size",
            "output_size",
            "num_layers",
            "dropout",
        ])


@encoder_register("lstm")
class LSTM(SimpleModule):
    def __init__(self, config: LSTMConfig):
        super(LSTM, self).__init__(config)
        self._provide_keys = {'embedding'}
        self._required_keys = {'embedding', 'attention_mask'}
        self._provided_keys = set()
        self.config = config
        self.lstm = module_register.get('lstm')(module_config_register.get('lstm')(config.lstm_config))

    def init_weight(self, method: Callable):
        """init  Module weight by `method`
        :method: init method
        :returns: None
        """
        for module in self.lstm.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """
        """
        inputs[self.get_output_name('embedding')] = self.lstm(inputs[self.get_input_name('embedding')], inputs[self.get_input_name('attention_mask')])
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('embedding')]]))
        return inputs



import torch
import torch.nn as nn
from typing import Dict
from . import module_register, module_config_register
from dlk.utils.config import BaseConfig

@module_config_register("crf")
class CRFConfig(BaseConfig):
    """docstring for LinearConfig
    {
        "config": {
            "output_size": 2,
            "batch_first": true,
            "reduction": "mean", //none|sum|mean|token_mean
        },
        "_name": "crf",
    }
    """
    def __init__(self, config: Dict):
        super(CRFConfig, self).__init__(config)
        config = config['config']
        self.output_size = config['output_size']
        if self.output_size <= 0:
            raise ValueError(f'invalid number of tags: {self.output_size}')
        self.post_check(config, used=[
            "output_size",
            "batch_first",
            "reduction",
        ])


@module_register("crf")
class ConditionalRandomField(nn.Module):
    r"""
    CRF, training_step for training, forward for decode。
    """

    def __init__(self, config: CRFConfig
                 ):
        super(ConditionalRandomField, self).__init__()

        self.num_tags = config.output_size

        self.transitions = nn.parameter.Parameter(torch.randn(self.num_tags, self.num_tags))
        self.start_transitions = nn.parameter.Parameter(torch.randn(self.num_tags))
        self.end_transitions = nn.parameter.Parameter(torch.randn(self.num_tags))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize the transition parameters.
        The parameters will be initialized randomly from a uniform distribution
        between -0.1 and 0.1.
        """
        nn.init.normal_(self.transitions, -1, 0.1)
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)

    def _normalizer_likelihood(self, logits, mask):
        r"""Computes the (batch_size,) denominator term for the log-likelihood, which is the
        sum of the likelihoods across all possible state sequences.
        :param logits:FloatTensor, max_len x batch_size x num_tags
        :param mask:ByteTensor, max_len x batch_size
        :return:FloatTensor, batch_size
        """
        seq_len, batch_size, n_tags = logits.size()
        alpha = logits[0]
        alpha = alpha + self.start_transitions.view(1, -1)

        flip_mask = mask.eq(False)

        for i in range(1, seq_len):
            emit_score = logits[i].view(batch_size, 1, n_tags)
            trans_score = self.transitions.view(1, n_tags, n_tags)
            tmp = alpha.view(batch_size, n_tags, 1) + emit_score + trans_score
            alpha = torch.logsumexp(tmp, 1).masked_fill(flip_mask[i].view(batch_size, 1), 0) + \
                    alpha.masked_fill(mask[i].eq(True).view(batch_size, 1), 0)

        alpha = alpha + self.end_transitions.view(1, -1)

        return torch.logsumexp(alpha, 1)

    def _gold_score(self, logits, tags, mask):
        r""" Compute the score for the gold path.
        :param logits: FloatTensor, max_len x batch_size x num_tags
        :param tags: LongTensor, max_len x batch_size
        :param mask: ByteTensor, max_len x batch_size
        :return:FloatTensor, batch_size
        """
        seq_len, batch_size, _ = logits.size()
        batch_idx = torch.arange(batch_size, dtype=torch.long, device=logits.device)
        seq_idx = torch.arange(seq_len, dtype=torch.long, device=logits.device)

        # trans_socre [L-1, B]
        mask = mask.eq(True)
        flip_mask = mask.eq(False)
        trans_score = self.transitions[tags[:seq_len - 1], tags[1:]].masked_fill(flip_mask[1:, :], 0)
        # emit_score [L, B]
        emit_score = logits[seq_idx.view(-1, 1), batch_idx.view(1, -1), tags].masked_fill(flip_mask, 0)
        # score [L-1, B]
        score = trans_score + emit_score[:seq_len - 1, :]
        score = score.sum(0) + emit_score[-1].masked_fill(flip_mask[-1], 0)
        st_scores = self.start_transitions.view(1, -1).repeat(batch_size, 1)[batch_idx, tags[0]]
        last_idx = mask.long().sum(0) - 1
        ed_scores = self.end_transitions.view(1, -1).repeat(batch_size, 1)[batch_idx, tags[last_idx, batch_idx]]
        score = score + st_scores + ed_scores
        # return [B,]
        return score

    def training_step(self, logits, tags, mask):
        r"""
        :param torch.FloatTensor logits->emissions: batch_size x max_len x num_tags
        :param torch.LongTensor tags: batch_size x max_len
        :param torch.ByteTensor mask: batch_size x max_len, mask==0 means padding
        :return: torch.FloatTensor, (batch_size,)
        """
        logits = logits.transpose(0, 1)
        tags = tags.transpose(0, 1).long()
        mask = mask.transpose(0, 1).float()
        all_path_score = self._normalizer_likelihood(logits, mask)
        gold_path_score = self._gold_score(logits, tags, mask)
        loss = all_path_score - gold_path_score
        return loss.mean()

    def forward(self, logits, mask):
        r"""predict
        :param torch.FloatTensor logits->emissions: batch_size x max_len x num_tags
        :param torch.LongTensor mask: batch_size x max_len, mask set to 0 means padding
        :return: torch.LongTensor batch_size x max_len
        """
        logits = logits.transpose(0, 1)  # L, B, H
        mask = mask.transpose(0, 1)
        return self._viterbi_decode(logits, mask)

    def _viterbi_decode(self, emissions: torch.FloatTensor,
                        mask: torch.LongTensor) -> torch.Tensor:
        # emissions: (seq_length, batch_size, num_tags)
        # mask: (seq_length, batch_size)
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        assert mask[0].all()

        seq_length, batch_size = mask.shape

        # Start transition and first emission
        # shape: (batch_size, num_tags)
        score = self.start_transitions + emissions[0]
        history = []

        # score is a tensor of size (batch_size, num_tags) where for every batch,
        # value at column j stores the score of the best tag sequence so far that ends
        # with tag j
        # history saves where the best tags candidate transitioned from; this is used
        # when we trace back the best tag sequence

        # Viterbi algorithm recursive case: we compute the score of the best tag sequence
        # for every possible next tag
        for i in range(1, seq_length):
            # Broadcast viterbi score for every possible next tag
            # shape: (batch_size, num_tags, 1)
            broadcast_score = score.unsqueeze(2)

            # Broadcast emission score for every possible current tag
            # shape: (batch_size, 1, num_tags)
            broadcast_emission = emissions[i].unsqueeze(1)

            # Compute the score tensor of size (batch_size, num_tags, num_tags) where
            # for each sample, entry at row i and column j stores the score of the best
            # tag sequence so far that ends with transitioning from tag i to tag j and emitting
            # shape: (batch_size, num_tags, num_tags)
            next_score = broadcast_score + self.transitions + broadcast_emission

            # Find the maximum score over all possible current tag
            # shape: (batch_size, num_tags)
            next_score, indices = next_score.max(dim=1)

            # Set score to the next score if this timestep is valid (mask == 1)
            # and save the index that produces the next score
            # shape: (batch_size, num_tags)
            score = torch.where(mask.bool()[i].unsqueeze(1), next_score, score)
            history.append(indices)

        # End transition score
        # shape: (batch_size, num_tags)
        score += self.end_transitions

        # Now, compute the best path for each sample

        # shape: (batch_size,)
        seq_ends = mask.long().sum(dim=0) - 1
        best_tags_list = []

        for idx in range(batch_size):
            # Find the tag which maximizes the score at the last timestep; this is our best tag
            # for the last timestep
            _, best_last_tag = score[idx].max(dim=0)
            best_tags = [best_last_tag.item()]

            # We trace back where the best last tag comes from, append that to our best tag
            # sequence, and trace it back again, and so on
            for hist in reversed(history[:seq_ends[idx]]):
                best_last_tag = hist[idx][best_tags[-1]]
                best_tags.append(best_last_tag.item())

            # Reverse the order because we start from the last timestep
            best_tags.reverse()
            best_tags_list.append(best_tags)

        output = []
        for tag_list in best_tags_list:
            if len(tag_list)<seq_length:
                tag_list = tag_list + [-1]*(seq_length-len(tag_list))
            output.append(tag_list)
        return torch.tensor(output, dtype=torch.long, device=mask.device)



"""
Loader the data from dict and generator DataFrame
"""
from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
import pandas as pd
from dlk.utils.logger import Logger

logger = Logger.get_logger()

@subprocessor_config_register('txt_reg_loader')
class TxtRegLoaderConfig(BaseConfig):
    """docstring for TxtRegLoaderConfig
        {
            "_name": "txt_reg_loader",
            "config": {
                "train":{ 
                    "data_set": {                   // for different stage, this processor will process different part of data
                        "train": ['train', 'valid', 'test', 'predict'],
                        "predict": ['predict'],
                        "online": ['online']
                    },
                    "input_map": {   // without necessery don't change this
                        "sentence": "sentence", //for single
                        "sentence_a": "sentence_a",  // for pair
                        "sentence_b": "sentence_b",
                        "uuid": "uuid",
                        "values": "values",
                    },
                    "output_map": {   // without necessery don't change this
                        "sentence": "sentence", //for single
                        "sentence_a": "sentence_a", //for pair
                        "sentence_b": "sentence_b",
                        "uuid": "uuid",
                        "values": "values",
                    },
                    "data_type": "single", // single or pair
                },
                "predict": "train",
                "online": "train",
            }
        }
        NOTE: the txt_regformat input is
        {
            "uuid": '**-**-**-**'
            "sentence": "I have an apple",
            "values":  [0.0]
        }
    """
    def __init__(self, stage, config: Dict):

        super(TxtRegLoaderConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.output_map = self.config['output_map']
        self.input_map = self.config['input_map']
        self.data_type = self.config['data_type']
        self.post_check(self.config, used=[
            "data_set",
            "output_map",
            "input_map",
            "data_type",
        ])


@subprocessor_register('txt_reg_loader')
class TxtRegLoader(ISubProcessor):
    """docstring for TxtRegLoader
    """

    def __init__(self, stage: str, config: TxtRegLoaderConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        self.data_type = config.data_type
        assert self.data_type in {'single', 'pair'}
        if not self.data_set:
            logger.info(f"Skip 'txt_reg_loader' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        '''
            data: {
                "train": list of json format train data
            }
        '''

        if not self.data_set:
            return data

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do txt_reg_loader on it.')
                continue
            data_set = data['data'][data_set_name]

            uuids = []
            valueses = []
            # for pair
            sentences_a = []
            sentences_b = []
            # for single
            sentences = []

            for one_ins in data_set:
                try:
                    if self.data_type == 'pair':
                        sentences_a.append(one_ins[self.config.input_map['sentence_a']])
                        sentences_b.append(one_ins[self.config.input_map['sentence_b']])
                    else:
                        sentences.append(one_ins[self.config.input_map['sentence']])
                    uuids.append(one_ins[self.config.input_map['uuid']])
                    valueses.append(one_ins[self.config.input_map['values']])
                except:
                    raise PermissionError(f"You must provide the data as requests, we need 'sentence', 'uuid' and 'values', or you can provide the input_map to map the origin data to this format")
            if self.data_type == 'pair':
                data_df = pd.DataFrame(data= {
                    self.config.output_map["sentence_a"]: sentences_a,
                    self.config.output_map["sentence_b"]: sentences_b,
                    self.config.output_map["uuid"]: uuids,
                    self.config.output_map["values"]: valueses,
                })
            else:
                data_df = pd.DataFrame(data= {
                    self.config.output_map["sentence"]: sentences,
                    self.config.output_map["uuid"]: uuids,
                    self.config.output_map["values"]: valueses,
                })
            data['data'][data_set_name] = data_df

        return data
